#!/usr/bin/env python3
#
# Tests the loading/saveing mmt files and states.
# See also test_parsing.py
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import numpy as np

import myokit

from myokit.tests import DIR_DATA, TemporaryDirectory


class LoadSaveMmtTest(unittest.TestCase):
    """Tests load/save functions for mmt files."""

    def test_examplify(self):
        # Test _examplify.
        self.assertEqual(myokit._io._examplify('test.txt'), 'test.txt')
        self.assertEqual(myokit._io._examplify('example'), myokit.EXAMPLE)

    def test_load_multiline_string_indent(self):
        # Test what happens when you load save a string that gets auto-indented

        # Create model with multi-line meta-data property
        d1 = 'First line\n\nSecond line'
        m1 = myokit.Model()
        m1.meta['desc'] = d1
        e = m1.add_component('engine')
        v = e.add_variable('time')
        v.set_binding('time')
        v.set_rhs(0)
        # Store to disk
        with TemporaryDirectory() as d:
            opath = d.path('multiline.mmt')
            myokit.save_model(opath, m1)
            # Load and compare the meta-data string
            m2 = myokit.load_model(opath)
            d2 = m2.meta['desc']
            self.assertEqual(d1, d2)
        # Create model with indented multi-line meta-data property
        d1 = '  First line\n\n  Second line'
        dr = 'First line\n\nSecond line'
        m1 = myokit.Model()
        m1.meta['desc'] = d1
        e = m1.add_component('engine')
        v = e.add_variable('time')
        v.set_binding('time')
        v.set_rhs(0)
        # Store to disk
        with TemporaryDirectory() as d:
            opath = d.path('multiline.mmt')
            myokit.save_model(opath, m1)
            # Load and compare the meta-data string
            m2 = myokit.load_model(opath)
            d2 = m2.meta['desc']
            self.assertEqual(d2, dr)
        # Create model with strangely indented multi-line meta-data property
        d1 = '  First line\n\n   Second line'
        dr = 'First line\n\n Second line'
        m1 = myokit.Model()
        m1.meta['desc'] = d1
        e = m1.add_component('engine')
        v = e.add_variable('time')
        v.set_binding('time')
        v.set_rhs(0)
        # Store to disk
        with TemporaryDirectory() as d:
            opath = d.path('multiline.mmt')
            myokit.save_model(opath, m1)
            # Load and compare the meta-data string
            m2 = myokit.load_model(opath)
            d2 = m2.meta['desc']
            self.assertEqual(d2, dr)

    def test_load_partial(self):
        # Test loading of partial files.
        mpath = os.path.join(DIR_DATA, 'beeler-1977-model.mmt')
        ppath = os.path.join(DIR_DATA, 'beeler-1977-protocol.mmt')
        spath = os.path.join(DIR_DATA, 'beeler-1977-script.mmt')

        m, p, x = myokit.load(mpath)
        self.assertIsInstance(m, myokit.Model)
        self.assertIsNone(p)
        self.assertIsNone(x)

        m, p, x = myokit.load(ppath)
        self.assertIsNone(m)
        self.assertIsInstance(p, myokit.Protocol)
        self.assertIsNone(x)

        m, p, x = myokit.load(spath)
        self.assertIsNone(m)
        self.assertIsNone(p)
        self.assertTrue(isinstance(x, str))

        self.assertRaises(
            myokit.SectionNotFoundError, myokit.load_model, ppath)
        self.assertRaises(
            myokit.SectionNotFoundError, myokit.load_model, spath)
        self.assertRaises(
            myokit.SectionNotFoundError, myokit.load_protocol, mpath)
        self.assertRaises(
            myokit.SectionNotFoundError, myokit.load_protocol, spath)
        self.assertRaises(
            myokit.SectionNotFoundError, myokit.load_script, mpath)
        self.assertRaises(
            myokit.SectionNotFoundError, myokit.load_script, ppath)

    def test_save(self):
        # Test if the correct parts are saved/loaded from disk using the
        # ``save()`` method.

        # Test example loading
        m, p, x = myokit.load('example')
        self.assertIsInstance(m, myokit.Model)
        self.assertIsInstance(p, myokit.Protocol)
        self.assertTrue(isinstance(x, str))

        # Save all three and reload
        with TemporaryDirectory() as d:
            opath = d.path('test.mmt')
            myokit.save(opath, m, p, x)
            mm, pp, xx = myokit.load(opath)
            self.assertEqual(m.code(), mm.code())
            self.assertEqual(p.code(), pp.code())
            self.assertEqual(x, xx)

        # Save only model
        with TemporaryDirectory() as d:
            opath = d.path('test.mmt')
            myokit.save(opath, model=m)
            with open(opath, 'r') as f:
                text = f.read()
            self.assertTrue('[[model]]' in text)
            self.assertFalse('[[protocol]]' in text)
            self.assertFalse('[[script]]' in text)
            mm, pp, xx = myokit.load(opath)
            self.assertEqual(mm.code(), m.code())
            self.assertEqual(pp, None)
            self.assertEqual(xx, None)

        # Save only protocol
        with TemporaryDirectory() as d:
            opath = d.path('test.mmt')
            myokit.save(opath, protocol=p)
            with open(opath, 'r') as f:
                text = f.read()
            self.assertFalse('[[model]]' in text)
            self.assertTrue('[[protocol]]' in text)
            self.assertFalse('[[script]]' in text)
            mm, pp, xx = myokit.load(opath)
            self.assertEqual(mm, None)
            self.assertEqual(pp.code(), p.code())
            self.assertEqual(xx, None)

        # Save only script
        with TemporaryDirectory() as d:
            opath = d.path('test.mmt')
            myokit.save(opath, script=x)
            with open(opath, 'r') as f:
                text = f.read()
            self.assertFalse('[[model]]' in text)
            self.assertFalse('[[protocol]]' in text)
            self.assertTrue('[[script]]' in text)
            mm, pp, xx = myokit.load(opath)
            self.assertEqual(mm, None)
            self.assertEqual(pp, None)
            self.assertEqual(xx, x)

        # Save all but model
        with TemporaryDirectory() as d:
            opath = d.path('test.mmt')
            myokit.save(opath, protocol=p, script=x)
            with open(opath, 'r') as f:
                text = f.read()
            self.assertFalse('[[model]]' in text)
            self.assertTrue('[[protocol]]' in text)
            self.assertTrue('[[script]]' in text)
            mm, pp, xx = myokit.load(opath)
            self.assertEqual(mm, None)
            self.assertEqual(pp.code(), p.code())
            self.assertEqual(xx, x)

        # Save all but protocol
        with TemporaryDirectory() as d:
            opath = d.path('test.mmt')
            myokit.save(opath, model=m, script=x)
            with open(opath, 'r') as f:
                text = f.read()
            self.assertTrue('[[model]]' in text)
            self.assertFalse('[[protocol]]' in text)
            self.assertTrue('[[script]]' in text)
            mm, pp, xx = myokit.load(opath)
            self.assertEqual(mm.code(), m.code())
            self.assertEqual(pp, None)
            self.assertEqual(xx, x)

        # Save all but script
        with TemporaryDirectory() as d:
            opath = d.path('test.mmt')
            myokit.save(opath, model=m, protocol=p)
            with open(opath, 'r') as f:
                text = f.read()
            self.assertTrue('[[model]]' in text)
            self.assertTrue('[[protocol]]' in text)
            self.assertFalse('[[script]]' in text)
            mm, pp, xx = myokit.load(opath)
            self.assertEqual(mm.code(), m.code())
            self.assertEqual(pp.code(), p.code())
            self.assertEqual(xx, None)

        # Save all as strings
        with TemporaryDirectory() as d:
            opath = d.path('test.mmt')
            myokit.save(opath, m, p, x)
            with open(opath, 'r') as f:
                text1 = f.read()
            myokit.save(opath, m.code(), p.code(), x)
            with open(opath, 'r') as f:
                text2 = f.read()
            self.assertEqual(text1, text2)

        # Save all as strings without [[model]] or [[protocol]] tage
        with TemporaryDirectory() as d:
            opath = d.path('test.mmt')
            myokit.save(opath, m, p, x)
            with open(opath, 'r') as f:
                text1 = f.read()
            mcode = '\n'.join(m.code().splitlines()[1:])
            pcode = '\n'.join(p.code().splitlines()[1:])
            myokit.save(opath, mcode, pcode, x)
            with open(opath, 'r') as f:
                text2 = f.read()
            self.assertEqual(text1, text2)

        # Save all, compare with string generated version
        with TemporaryDirectory() as d:
            opath = d.path('test.mmt')
            myokit.save(opath, model=m, protocol=p, script=x)
            with open(opath, 'r') as f:
                text = f.read()
            self.assertEqual(text, myokit.save(model=m, protocol=p, script=x))

        # Save nothing
        with TemporaryDirectory() as d:
            opath = d.path('test.mmt')
            self.assertRaises(ValueError, myokit.save, opath)
            self.assertFalse(os.path.exists(opath))

    def test_save_model(self):
        # Test if the correct parts are saved/loaded from disk using the
        # ``save_model()`` method.

        ipath = os.path.join(DIR_DATA, 'lr-1991.mmt')
        # Test example loading
        m = myokit.load_model('example')
        self.assertIsInstance(m, myokit.Model)
        # Test file loading
        m = myokit.load_model(ipath)
        self.assertIsInstance(m, myokit.Model)
        with TemporaryDirectory() as d:
            opath = d.path('loadsave.mmt')
            myokit.save_model(opath, m)
            # Test no other parts were written
            with open(opath, 'r') as f:
                text = f.read()
            self.assertTrue('[[model]]' in text)
            self.assertFalse('[[protocol]]' in text)
            self.assertFalse('[[script]]' in text)
            # Test reloading
            mm = myokit.load_model(opath)
            self.assertIsInstance(mm, myokit.Model)
            self.assertEqual(mm.code(), m.code())

    def test_save_protocol(self):
        # Test if the correct parts are saved/loaded from disk using the
        # ``save_protocol()`` method.

        ipath = os.path.join(DIR_DATA, 'lr-1991.mmt')
        # Test example loading
        p = myokit.load_protocol('example')
        self.assertIsInstance(p, myokit.Protocol)
        # Test file loading
        p = myokit.load_protocol(ipath)
        self.assertIsInstance(p, myokit.Protocol)
        with TemporaryDirectory() as d:
            opath = d.path('test.mmt')
            myokit.save_protocol(opath, p)
            # Test no other parts were written
            with open(opath, 'r') as f:
                text = f.read()
            self.assertFalse('[[model]]' in text)
            self.assertTrue('[[protocol]]' in text)
            self.assertFalse('[[script]]' in text)
            # Test reloading
            pp = myokit.load_protocol(opath)
            self.assertIsInstance(pp, myokit.Protocol)
            self.assertEqual(pp.code(), p.code())

    def test_save_script(self):
        # Test if the correct parts are saved/loaded from disk using the
        # ``save_script()`` method.

        ipath = os.path.join(DIR_DATA, 'lr-1991.mmt')
        # Test example loading
        x = myokit.load_script('example')
        self.assertTrue(isinstance(x, str))
        # Test file loading
        x = myokit.load_script(ipath)
        self.assertTrue(isinstance(x, str))
        with TemporaryDirectory() as d:
            opath = d.path('test.mmt')
            myokit.save_script(opath, x)
            # Test no other parts were written
            with open(opath, 'r') as f:
                text = f.read()
            self.assertFalse('[[model]]' in text)
            self.assertFalse('[[protocol]]' in text)
            self.assertTrue('[[script]]' in text)
            # Test reloading
            xx = myokit.load_script(opath)
            self.assertTrue(isinstance(xx, str))
            self.assertEqual(x, xx)


class LoadSaveStateTest(unittest.TestCase):
    """Tests loading and saving states."""

    def test_load_save_state(self):
        # Test loading/saving state.
        m, p, x = myokit.load('example')
        with TemporaryDirectory() as d:
            # Test save and load without model
            f = d.path('state.txt')
            myokit.save_state(f, m.initial_values(True))
            self.assertEqual(myokit.load_state(f), m.initial_values(True))

            # Test save and load with model argument
            myokit.save_state(f, m.initial_values(True), m)
            self.assertEqual(myokit.load_state(f, m), m.initial_values(True))

            # Save without, load with model
            myokit.save_state(f, m.initial_values(True))
            self.assertEqual(myokit.load_state(f, m), m.initial_values(True))

            # Save with model, load without
            # Loaded version is dict!
            myokit.save_state(f, m.initial_values(True), m)
            self.assertEqual(myokit.load_state(f), dict(zip(
                [v.qname() for v in m.states()], m.initial_values(True))))

    def test_load_save_state_bin(self):
        # Test loading/saving state in binary format.
        m, p, x = myokit.load('example')
        with TemporaryDirectory() as d:

            # Test save and load with double precision
            f = d.path('state.bin')
            myokit.save_state_bin(f, m.initial_values(True))
            self.assertEqual(myokit.load_state_bin(f), m.initial_values(True))

            # Test save and load with single precision
            f = d.path('state.bin')
            myokit.save_state_bin(
                f, m.initial_values(True), myokit.SINGLE_PRECISION)
            d = np.array(myokit.load_state_bin(f))
            d -= np.array(m.initial_values(True))
            self.assertTrue(np.all(np.abs(d) < 1e-5))   # Not very precise!


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests the DataLog class
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import numpy as np

import myokit

from myokit.tests import (
    DIR_DATA,
    DIR_IO,
    TemporaryDirectory,
    TestReporter,
    CancellingReporter,
    WarningCollector,
)

# Extra output
debug = False


class DataLogTest(unittest.TestCase):
    """
    Tests the DataLog's functions.
    """

    def test_extend(self):
        # Test the extend function.

        d1 = myokit.DataLog(time='time')
        v1 = [10, 9, 10, -1, -1, -1]
        v2 = [12, 2, 43, 31, 2, 7]
        d1['time'] = [1, 2, 3]
        d1['v1'] = v1[:3]
        d1['v2'] = v2[:3]
        d2 = myokit.DataLog()
        d2.set_time_key('time')
        d2['time'] = [4, 5, 6]
        d2['v1'] = v1[3:]
        d2['v2'] = v2[3:]

        # No errors, method should work
        d3 = d1.extend(d2)
        self.assertEqual(d3.time(), [1, 2, 3, 4, 5, 6])
        self.assertEqual(d3['v1'], v1)
        self.assertEqual(d3['v2'], v2)

        # Both logs must have the same time key
        d2.set_time_key(None)
        self.assertRaises(ValueError, d1.extend, d2)
        d2.set_time_key('time')
        d1.extend(d2)
        d2.set_time_key('v1')
        self.assertRaises(ValueError, d1.extend, d2)
        d2.set_time_key('time')

        # Both logs must have the same keys
        d2['v3'] = v1
        self.assertRaises(ValueError, d1.extend, d2)
        del d2['v3']

        # Times can overlap
        d2['time'] = [3, 4, 5]
        d1.extend(d2)
        # But times in d2 must be >= than times in d1
        d2['time'] = [2, 3, 4]
        self.assertRaises(ValueError, d1.extend, d2)
        d2['time'] = [4, 5, 6]

        # Should work with two numpy arrays
        d3 = d1.npview().extend(d2.npview())
        self.assertEqual(list(d3.time()), [1, 2, 3, 4, 5, 6])
        self.assertEqual(list(d3['v1']), v1)
        self.assertEqual(list(d3['v2']), v2)

        # Should work with one numpy array
        d3 = d1.extend(d2.npview())
        self.assertEqual(list(d3.time()), [1, 2, 3, 4, 5, 6])
        self.assertEqual(list(d3['v1']), v1)
        self.assertEqual(list(d3['v2']), v2)

        # Should work with partial numpy arrays
        d1['v1'] = np.array(v1[:3])
        d2['v2'] = np.array(v2[3:])
        self.assertIsInstance(d1.time(), list)
        self.assertIsInstance(d1['v1'], np.ndarray)
        self.assertIsInstance(d1['v2'], list)
        self.assertIsInstance(d2.time(), list)
        self.assertIsInstance(d2['v1'], list)
        self.assertIsInstance(d2['v2'], np.ndarray)
        d3 = d1.extend(d2)
        self.assertIsInstance(d3.time(), list)
        self.assertIsInstance(d3['v1'], np.ndarray)
        self.assertIsInstance(d3['v2'], np.ndarray)
        self.assertEqual(list(d3.time()), [1, 2, 3, 4, 5, 6])
        self.assertEqual(list(d3['v1']), v1)
        self.assertEqual(list(d3['v2']), v2)

    def test_find(self):
        # Tests the deprecated find() function

        d = myokit.DataLog({
            'engine.time': [0, 5, 10, 15, 20],
            'membrane.V': [0, 50, 100, 150, 200]})
        d.set_time_key('engine.time')
        with WarningCollector() as w:
            self.assertEqual(d.find(-5), d.find_after(-5))
        self.assertIn('eprecated', w.text())
        with WarningCollector():
            self.assertEqual(d.find(0), d.find_after(0))
            self.assertEqual(d.find(2), d.find_after(2))
            self.assertEqual(d.find(5), d.find_after(5))
            self.assertEqual(d.find(7), d.find_after(7))
            self.assertEqual(d.find(20), d.find_after(20))
            self.assertEqual(d.find(22), d.find_after(22))

    def test_find_after(self):
        # Test the find_after() function.

        x = myokit.DataLog({
            'engine.time': [0, 5, 10, 15, 20],
            'membrane.V': [0, 50, 100, 150, 200]})
        x.set_time_key('engine.time')
        self.assertEqual(x.find_after(-5), 0)
        self.assertEqual(x.find_after(0), 0)
        self.assertEqual(x.find_after(2), 1)
        self.assertEqual(x.find_after(5), 1)
        self.assertEqual(x.find_after(8), 2)
        self.assertEqual(x.find_after(10), 2)
        self.assertEqual(x.find_after(13), 3)
        self.assertEqual(x.find_after(15), 3)
        self.assertEqual(x.find_after(19), 4)
        self.assertEqual(x.find_after(20), 4)
        self.assertEqual(x.find_after(21), 5)

        # Now with a numpy log
        x = x.npview()
        self.assertEqual(x.find_after(-5), 0)
        self.assertEqual(x.find_after(0), 0)
        self.assertEqual(x.find_after(2), 1)
        self.assertEqual(x.find_after(5), 1)
        self.assertEqual(x.find_after(8), 2)
        self.assertEqual(x.find_after(10), 2)
        self.assertEqual(x.find_after(13), 3)
        self.assertEqual(x.find_after(15), 3)
        self.assertEqual(x.find_after(19), 4)
        self.assertEqual(x.find_after(20), 4)
        self.assertEqual(x.find_after(21), 5)

    def test_indexing(self):
        # Test the indexing overrides in the simulation log.

        d = myokit.DataLog()
        d['x'] = 'y'
        self.assertEqual(d['x'], 'y')
        d['1.2.membrane.V'] = 'a bear'
        x = d['1.2.membrane.V']
        y = d['membrane.V', 1, 2]
        z = d['membrane.V', (1, 2)]
        self.assertEqual(x, y, z)
        d['4.a'] = 5
        self.assertEqual(d['4.a'], 5)
        d['a', 4] = 6
        self.assertEqual(d['a', 4], 6)
        d['a', 4, 5] = 7
        self.assertEqual(d['a', 4, 5], 7)
        d['a', (5, 6)] = 8
        self.assertEqual(d['a', 5, 6], 8)
        self.assertTrue('4.a' in d)
        self.assertTrue(('a', 4) in d)
        self.assertFalse('5.a' in d)
        self.assertFalse(('a', 5) in d)
        self.assertTrue(('a', (5, 6)) in d)
        del d['a', (5, 6)]
        self.assertFalse(('a', (5, 6)) in d)

    def test_integrate(self):
        # Test the integrate method.

        # Create an irregular time array
        from random import random
        t = []
        for i in range(0, 100):
            t.append(5 * random())
        t.sort()
        t = np.array(t)

        # Create a signal and its derivative
        i = 15 * t**2 + 4 * t + 2
        q = 5 * t**3 + 2 * t**2 + 2 * t

        # Since q already has a value for q[0], remove it to look at the
        # error in integrating...
        q -= q[0]

        # Get estimated value from log
        d = myokit.DataLog({'engine.time': t, 'ina.INa': i})
        d.set_time_key('engine.time')
        x = d.integrate('ina.INa')

        # Get relative error
        e = np.abs(x[1:] - q[1:]) / q[1:]

        # Get mean
        e = np.mean(e)
        if debug:
            print(e)
        self.assertLess(e, 0.1)
        if debug:
            import matplotlib.pyplot as plt
            plt.figure()
            plt.plot(t, q, label='original')
            plt.plot(t, x, label='mid sum')
            plt.legend(loc='upper left')
            plt.figure()
            plt.plot(t, np.abs(x - q) / q, label='mid sum error')
            plt.legend(loc='upper left')
            plt.show()

    def test_interpolate_at(self):
        # Test the interpolate_at method.

        d = myokit.DataLog(time='t')
        d['t'] = [0, 1, 2, 3]
        d['v'] = [0, 10, 30, 60]
        self.assertEqual(d.interpolate_at('v', 0), 0)
        self.assertEqual(d.interpolate_at('v', 0.5), 5)
        self.assertEqual(d.interpolate_at('v', 1), 10)
        self.assertEqual(d.interpolate_at('v', 1.5), 20)
        self.assertEqual(d.interpolate_at('v', 2), 30)
        self.assertEqual(d.interpolate_at('v', 2.5), 45)
        self.assertEqual(d.interpolate_at('v', 3), 60)

        self.assertRaises(ValueError, d.interpolate_at, 'v', -0.5)
        self.assertRaises(ValueError, d.interpolate_at, 'v', 3 + 1e-9)

        d = myokit.DataLog(time='t')
        d['t'] = [2, 4]
        d['v'] = [200, 400]
        self.assertEqual(d.interpolate_at('v', 2), 200)
        self.assertEqual(d.interpolate_at('v', 2.4), 240)
        self.assertEqual(d.interpolate_at('v', 4), 400)

        self.assertRaises(ValueError, d.interpolate_at, 'v', 1)
        self.assertRaises(ValueError, d.interpolate_at, 'v', 5)

    def test_itrim(self):
        # Test the itrim() method.

        d = myokit.DataLog()
        d.set_time_key('t')
        d['t'] = t = [0, 0.1, 0.2, 0.3, 0.4]
        d['a'] = a = [1, 2, 3, 4, 5]
        d['b'] = b = [10, 20, 30, 40, 50]

        # Test function
        def tr(i, j):
            e = d.itrim(i, j)
            self.assertEqual(len(d['t']), 5)
            self.assertEqual(d['t'], t)
            self.assertEqual(d['a'], a)
            self.assertEqual(list(d['b']), list(b))
            self.assertEqual(e['t'], t[i:j])
            self.assertEqual(e['a'], a[i:j])
            self.assertEqual(list(e['b']), list(b[i:j]))

        # Normal operation
        tr(1, 3)
        tr(-5, 3)
        tr(1, 30)
        tr(-10, 40)

        # Partial numpy
        d['b'] = b = np.array(b)
        tr(1, 3)
        tr(-5, 3)
        tr(1, 30)
        tr(-10, 40)

    def test_itrim_left(self):
        # Test the itrim_left() method.

        d = myokit.DataLog()
        d.set_time_key('t')
        d['t'] = t = [0, 0.1, 0.2, 0.3, 0.4]
        d['a'] = a = [1, 2, 3, 4, 5]
        d['b'] = b = [10, 20, 30, 40, 50]

        # Test function
        def tr(i):
            e = d.itrim_left(i)
            self.assertEqual(len(d['t']), 5)
            self.assertEqual(d['t'], t)
            self.assertEqual(d['a'], a)
            self.assertEqual(list(d['b']), list(b))
            self.assertEqual(e['t'], t[i:])
            self.assertEqual(e['a'], a[i:])
            self.assertEqual(list(e['b']), list(b[i:]))

        # Normal operation
        tr(2)
        tr(-5)
        tr(30)

        # Partial numpy
        d['b'] = b = np.array(b)
        tr(2)
        tr(-5)
        tr(30)

    def test_itrim_right(self):
        # Test the itrim_right() method.

        d = myokit.DataLog()
        d.set_time_key('t')
        d['t'] = t = [0, 0.1, 0.2, 0.3, 0.4]
        d['a'] = a = [1, 2, 3, 4, 5]
        d['b'] = b = [10, 20, 30, 40, 50]

        # Test function
        def tr(i):
            e = d.itrim_right(i)
            self.assertEqual(len(d['t']), 5)
            self.assertEqual(d['t'], t)
            self.assertEqual(d['a'], a)
            self.assertEqual(list(d['b']), list(b))
            self.assertEqual(e['t'], t[:i])
            self.assertEqual(e['a'], a[:i])
            self.assertEqual(list(e['b']), list(b[:i]))

        # Normal operation
        tr(2)
        tr(-5)
        tr(30)

        # Partial numpy
        d['b'] = b = np.array(b)
        tr(2)
        tr(-5)
        tr(30)

    def test_keys_like(self):
        # Test the keys_like(query) method.

        d = myokit.DataLog()
        d.set_time_key('t')
        d['t'] = [1, 2, 3, 4]
        v = [0, 0, 0, 0]

        # Zero dimensional variable
        e = d.clone()
        e['v'] = v
        self.assertEqual(e.keys_like('v'), [])

        # Non-existant variable
        self.assertEqual(e.keys_like('w'), [])

        # 1-d
        e = d.clone()
        e['0.v'] = e['1.v'] = e['2.v'] = v
        self.assertEqual(e.keys_like('v'), ['0.v', '1.v', '2.v'])
        e['0.m.v'] = e['1.m.v'] = e['2.m.v'] = v
        self.assertEqual(e.keys_like('m.v'), ['0.m.v', '1.m.v', '2.m.v'])
        e['2.w'] = e['1.w'] = e['0.w'] = v
        self.assertEqual(e.keys_like('w'), ['0.w', '1.w', '2.w'])
        e['20.x'] = e['10.x'] = e['0.x'] = v
        self.assertEqual(e.keys_like('x'), ['0.x', '10.x', '20.x'])

        # 2-d
        e = d.clone()
        e['0.0.m.v'] = e['1.0.m.v'] = e['0.1.m.v'] = e['1.1.m.v'] = v
        self.assertEqual(
            e.keys_like('m.v'),
            ['0.0.m.v', '0.1.m.v', '1.0.m.v', '1.1.m.v'])

    def test_prepare_log_0d(self):
        # Test the `prepare_log` method for single-cell simulations.

        # Test multi-cell log preparing
        from myokit import prepare_log
        m = myokit.load_model(
            os.path.join(DIR_DATA, 'lr-1991-testing.mmt'))

        #
        # 1. Integer flags
        #

        # No variables
        d = prepare_log(myokit.LOG_NONE, m)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 0)

        # States
        d = prepare_log(myokit.LOG_STATE, m)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 8)
        self.assertIn('membrane.V', d)
        self.assertIn('ina.m', d)
        self.assertIn('ina.h', d)
        self.assertIn('ina.j', d)
        self.assertIn('ica.d', d)
        self.assertIn('ica.f', d)
        self.assertIn('ik.x', d)
        self.assertIn('ica.Ca_i', d)
        self.assertNotIn('garbage', d)
        self.assertNotIn('engine.time', d)

        # Internal variables
        d = prepare_log(myokit.LOG_INTER, m)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertIn('membrane.i_stim', d)
        self.assertIn('membrane.i_ion', d)
        self.assertIn('ina.a', d)
        self.assertIn('ina.m.alpha', d)
        self.assertIn('ina.m.beta', d)
        self.assertIn('ina.h.alpha', d)
        self.assertIn('ina.h.beta', d)
        self.assertIn('ina.j.alpha', d)
        self.assertIn('ina.j.beta', d)
        self.assertIn('ina.INa', d)
        self.assertIn('ik.xi', d)
        self.assertIn('ik.x.alpha', d)
        self.assertIn('ik.x.beta', d)
        self.assertIn('ik.IK', d)
        self.assertIn('ikp.IKp', d)
        self.assertIn('ica.E', d)
        self.assertIn('ica.d.alpha', d)
        self.assertIn('ica.d.beta', d)
        self.assertIn('ica.d.beta.va', d)
        self.assertIn('ica.d.beta.vb', d)
        self.assertIn('ica.d.beta.vc', d)
        self.assertIn('ica.d.beta.vc.vd', d)
        self.assertIn('ica.f.alpha', d)
        self.assertIn('ica.f.beta', d)
        self.assertIn('ica.ICa', d)
        self.assertIn('ica.ICa.nest1', d)
        self.assertIn('ica.ICa.nest2', d)
        self.assertIn('ik1.gK1', d)
        self.assertIn('ik1.gK1.alpha', d)
        self.assertIn('ik1.gK1.beta', d)
        self.assertIn('ik1.IK1', d)
        self.assertIn('ib.Ib', d)
        self.assertEqual(len(d), 32)

        # Bound variables
        d = prepare_log(myokit.LOG_BOUND, m)
        self.assertIn('engine.time', d)
        self.assertIn('engine.pace', d)
        self.assertIn('membrane.i_diff', d)
        self.assertEqual(len(d), 3)

        # Combinations
        self.assertEqual(
            len(prepare_log(myokit.LOG_NONE + myokit.LOG_BOUND, m)), 3)
        self.assertEqual(
            len(prepare_log(myokit.LOG_NONE + myokit.LOG_INTER, m)), 32)
        self.assertEqual(
            len(prepare_log(myokit.LOG_BOUND + myokit.LOG_STATE, m)), 11)
        self.assertEqual(
            len(prepare_log(myokit.LOG_STATE + myokit.LOG_INTER, m)), 40)

        #
        # 2. None
        #

        d = prepare_log(None, m, if_empty=myokit.LOG_NONE)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 0)
        d = prepare_log(None, m, if_empty=myokit.LOG_STATE)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 8)

        #
        # 3. List of names
        #

        d = prepare_log(('engine.time', 'membrane.V'), m)
        self.assertEqual(len(d), 2)
        d = prepare_log((
            'engine.time',
            'engine.pace',
            'membrane.V',
            'ina.m',
            'ina.h',
            'ina.j',
            'ica.d',
            'ica.f',
            'ik.x',
            'ica.Ca_i',
            'membrane.i_stim',
            'membrane.i_ion',
            'membrane.i_diff',
            'ina.a',
            'ina.m.alpha',
            'ina.m.beta',
            'ina.h.alpha',
            'ina.h.beta',
            'ina.j.alpha',
            'ina.j.beta',
            'ina.INa',
            'ik.xi',
            'ik.x.alpha',
            'ik.x.beta',
            'ik.IK',
            'ikp.IKp',
            'ica.E',
            'ica.d.alpha',
            'ica.d.beta',
            'ica.d.beta.va',
            'ica.d.beta.vb',
            'ica.d.beta.vc',
            'ica.d.beta.vc.vd',
            'ica.f.alpha',
            'ica.f.beta',
            'ica.ICa',
            'ica.ICa.nest1',
            'ica.ICa.nest2',
            'ik1.gK1',
            'ik1.gK1.alpha',
            'ik1.gK1.beta',
            'ik1.IK1',
            'ib.Ib',), m)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 32 + 8 + 3)
        self.assertIn('engine.time', d)
        self.assertIn('engine.pace', d)
        self.assertIn('membrane.V', d)
        self.assertIn('ina.m', d)
        self.assertIn('ina.h', d)
        self.assertIn('ina.j', d)
        self.assertIn('ica.d', d)
        self.assertIn('ica.f', d)
        self.assertIn('ik.x', d)
        self.assertIn('ica.Ca_i', d)
        self.assertIn('membrane.i_stim', d)
        self.assertIn('ina.a', d)
        self.assertIn('ina.m.alpha', d)
        self.assertIn('ina.m.beta', d)
        self.assertIn('ina.h.alpha', d)
        self.assertIn('ina.h.beta', d)
        self.assertIn('ina.j.alpha', d)
        self.assertIn('ina.j.beta', d)
        self.assertIn('ina.INa', d)
        self.assertIn('ik.xi', d)
        self.assertIn('ik.x.alpha', d)
        self.assertIn('ik.x.beta', d)
        self.assertIn('ik.IK', d)
        self.assertIn('ikp.IKp', d)
        self.assertIn('ica.E', d)
        self.assertIn('ica.d.alpha', d)
        self.assertIn('ica.d.beta', d)
        self.assertIn('ica.d.beta.va', d)
        self.assertIn('ica.d.beta.vb', d)
        self.assertIn('ica.d.beta.vc', d)
        self.assertIn('ica.d.beta.vc.vd', d)
        self.assertIn('ica.f.alpha', d)
        self.assertIn('ica.f.beta', d)
        self.assertIn('ica.ICa', d)
        self.assertIn('ica.ICa.nest1', d)
        self.assertIn('ica.ICa.nest2', d)
        self.assertIn('ik1.gK1', d)
        self.assertIn('ik1.gK1.alpha', d)
        self.assertIn('ik1.gK1.beta', d)
        self.assertIn('ik1.IK1', d)
        self.assertIn('ib.Ib', d)

        # Variables or LhsExpressions in list
        d = prepare_log([m.get('membrane.V')], m)
        self.assertIn('membrane.V', d)
        d = prepare_log([m.get('membrane.V').lhs()], m)
        self.assertIn('dot(membrane.V)', d)

        #
        # 4. Existing log
        #
        # Empty log
        d = prepare_log({}, m)
        self.assertEqual(len(d), 0)
        #
        d = prepare_log({'engine.time': []}, m)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 1)
        self.assertIn('engine.time', d)
        self.assertNotIn('membrane.V', d)
        d = prepare_log({'engine.time': [1, 2, 3]}, m)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 1)
        self.assertIn('engine.time', d)
        self.assertNotIn('membrane.V', d)
        self.assertEqual(len(d['engine.time']), 3)
        d = prepare_log({'engine.time': [1, 2, 3], 'membrane.V': [1, 2, 3]}, m)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 2)
        d = prepare_log(d, m)
        self.assertIsInstance(d, myokit.DataLog)
        d = prepare_log(d, m)
        self.assertIsInstance(d, myokit.DataLog)

        #
        # 5. Errors
        #

        # Disallowed variable types, and explicitly specified variables
        self.assertRaisesRegex(
            ValueError, 'support constants', prepare_log, ['cell.Na_o'], m)
        self.assertRaisesRegex(
            ValueError, 'support state', prepare_log, ['membrane.V'], m,
            allowed_classes=myokit.LOG_BOUND)
        self.assertRaisesRegex(
            ValueError, 'support bound', prepare_log, ['engine.pace'], m,
            allowed_classes=myokit.LOG_STATE)
        self.assertRaisesRegex(
            ValueError, 'support intermediary', prepare_log, ['ina.INa'], m,
            allowed_classes=myokit.LOG_STATE)

        # Empty log argument, but bad `if-empty` argument
        self.assertRaisesRegex(
            ValueError, 'if_empty', prepare_log, None, m,
            if_empty=myokit.LOG_STATE, allowed_classes=myokit.LOG_BOUND)

        # Disallowed variable types, and integer flags for variables
        self.assertRaisesRegex(
            ValueError, 'support state', prepare_log, myokit.LOG_STATE, m,
            allowed_classes=myokit.LOG_BOUND)
        self.assertRaisesRegex(
            ValueError, 'support bound', prepare_log, myokit.LOG_BOUND, m,
            allowed_classes=myokit.LOG_STATE)
        self.assertRaisesRegex(
            ValueError, 'support intermediary', prepare_log, myokit.LOG_INTER,
            m, allowed_classes=myokit.LOG_STATE)
        self.assertRaisesRegex(
            ValueError, 'support time-derivatives', prepare_log,
            myokit.LOG_DERIV, m, allowed_classes=myokit.LOG_STATE)

        # Unknown integer flags
        self.assertRaisesRegex(ValueError, 'unknown flag', prepare_log, -1, m)

        # Unknown variable in log/dict
        d = myokit.DataLog()
        d['bert.bert'] = []
        self.assertRaisesRegex(
            ValueError, 'Unknown variable', prepare_log, d, m)

        # Unsupported time-derivative in log/dict
        d = myokit.DataLog()
        d['dot(membrane.V)'] = []
        self.assertRaisesRegex(
            ValueError, 'derivatives', prepare_log, d, m,
            allowed_classes=myokit.LOG_STATE)

        # Time-derivative of non-state in log/dict
        d = myokit.DataLog()
        d['dot(ina.INa)'] = []
        self.assertRaisesRegex(
            ValueError, 'derivative of non-state', prepare_log, d, m)

        # Log given with objects that we can't append to
        self.assertRaisesRegex(
            ValueError, 'support the append', prepare_log,
            {'membrane.V': 'hi'}, m)

        # Argument `log` doesn't match any of the options
        self.assertRaisesRegex(
            ValueError, 'unexpected type', prepare_log, IOError, m)
        self.assertRaisesRegex(
            ValueError, 'String passed', prepare_log, 'membrane.V', m)

        # Unknown variable in list
        self.assertRaisesRegex(
            ValueError, 'Unknown variable', prepare_log, ['bert.bert'], m)

        # Unsupported time-derivative in list
        self.assertRaisesRegex(
            ValueError, 'derivatives', prepare_log, ['dot(membrane.V)'], m,
            allowed_classes=myokit.LOG_STATE)

        # Time-derivative of non-state in list
        self.assertRaisesRegex(
            ValueError, 'derivative of non-state', prepare_log,
            ['dot(ina.INa)'], m)

    def test_prepare_log_2d(self):
        # Test the `prepare_log` method for 2-dimensional logs.

        # Test multi-cell log preparing
        from myokit import prepare_log
        m = myokit.load_model(
            os.path.join(DIR_DATA, 'lr-1991-testing.mmt'))

        #
        # 1. Integer flags
        #

        # No variables
        d = prepare_log(myokit.LOG_NONE, m, (2, 1))
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 0)

        # States
        d = prepare_log(myokit.LOG_STATE, m, (2, 1))
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 16)
        self.assertIn('0.0.membrane.V', d)
        self.assertIn('0.0.ina.m', d)
        self.assertIn('0.0.ina.h', d)
        self.assertIn('0.0.ina.j', d)
        self.assertIn('0.0.ica.d', d)
        self.assertIn('0.0.ica.f', d)
        self.assertIn('0.0.ik.x', d)
        self.assertIn('0.0.ica.Ca_i', d)
        self.assertIn('1.0.membrane.V', d)
        self.assertIn('1.0.ina.m', d)
        self.assertIn('1.0.ina.h', d)
        self.assertIn('1.0.ina.j', d)
        self.assertIn('1.0.ica.d', d)
        self.assertIn('1.0.ica.f', d)
        self.assertIn('1.0.ik.x', d)
        self.assertIn('1.0.ica.Ca_i', d)
        self.assertNotIn('membrane.V', d)
        self.assertNotIn('ina.m', d)
        self.assertNotIn('ina.h', d)
        self.assertNotIn('ina.j', d)
        self.assertNotIn('ica.d', d)
        self.assertNotIn('ica.f', d)
        self.assertNotIn('ik.x', d)
        self.assertNotIn('ica.Ca_i', d)
        self.assertNotIn('garbage', d)
        self.assertNotIn('engine.time', d)
        d = prepare_log(myokit.LOG_STATE, m, (2, 2))
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 32)
        d = prepare_log(myokit.LOG_STATE, m, (4, 2))
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 64)

        # Internal variables
        d = prepare_log(myokit.LOG_INTER, m, (2, 1))
        self.assertIsInstance(d, myokit.DataLog)

        #self.assertEqual(len(d), 64)
        self.assertIn('0.0.membrane.i_stim', d)
        self.assertIn('0.0.membrane.i_ion', d)
        self.assertIn('0.0.ina.a', d)
        self.assertIn('0.0.ina.m.alpha', d)
        self.assertIn('0.0.ina.m.beta', d)
        self.assertIn('0.0.ina.h.alpha', d)
        self.assertIn('0.0.ina.h.beta', d)
        self.assertIn('0.0.ina.j.alpha', d)
        self.assertIn('0.0.ina.j.beta', d)
        self.assertIn('0.0.ina.INa', d)
        self.assertIn('0.0.ik.xi', d)
        self.assertIn('0.0.ik.x.alpha', d)
        self.assertIn('0.0.ik.x.beta', d)
        self.assertIn('0.0.ik.IK', d)
        self.assertIn('0.0.ikp.IKp', d)
        self.assertIn('0.0.ica.E', d)
        self.assertIn('0.0.ica.d.alpha', d)
        self.assertIn('0.0.ica.d.beta', d)
        self.assertIn('0.0.ica.d.beta.va', d)
        self.assertIn('0.0.ica.d.beta.vb', d)
        self.assertIn('0.0.ica.d.beta.vc', d)
        self.assertIn('0.0.ica.d.beta.vc.vd', d)
        self.assertIn('0.0.ica.f.alpha', d)
        self.assertIn('0.0.ica.f.beta', d)
        self.assertIn('0.0.ica.ICa', d)
        self.assertIn('0.0.ica.ICa.nest1', d)
        self.assertIn('0.0.ica.ICa.nest2', d)
        self.assertIn('0.0.ik1.gK1', d)
        self.assertIn('0.0.ik1.gK1.alpha', d)
        self.assertIn('0.0.ik1.gK1.beta', d)
        self.assertIn('0.0.ik1.IK1', d)
        self.assertIn('0.0.ib.Ib', d)
        self.assertIn('1.0.membrane.i_stim', d)
        self.assertIn('1.0.membrane.i_ion', d)
        self.assertIn('1.0.ina.a', d)
        self.assertIn('1.0.ina.m.alpha', d)
        self.assertIn('1.0.ina.m.beta', d)
        self.assertIn('1.0.ina.h.alpha', d)
        self.assertIn('1.0.ina.h.beta', d)
        self.assertIn('1.0.ina.j.alpha', d)
        self.assertIn('1.0.ina.j.beta', d)
        self.assertIn('1.0.ina.INa', d)
        self.assertIn('1.0.ik.xi', d)
        self.assertIn('1.0.ik.x.alpha', d)
        self.assertIn('1.0.ik.x.beta', d)
        self.assertIn('1.0.ik.IK', d)
        self.assertIn('1.0.ikp.IKp', d)
        self.assertIn('1.0.ica.E', d)
        self.assertIn('1.0.ica.d.alpha', d)
        self.assertIn('1.0.ica.d.beta', d)
        self.assertIn('1.0.ica.d.beta.va', d)
        self.assertIn('1.0.ica.d.beta.vb', d)
        self.assertIn('1.0.ica.d.beta.vc', d)
        self.assertIn('1.0.ica.d.beta.vc.vd', d)
        self.assertIn('1.0.ica.f.alpha', d)
        self.assertIn('1.0.ica.f.beta', d)
        self.assertIn('1.0.ica.ICa', d)
        self.assertIn('1.0.ica.ICa.nest1', d)
        self.assertIn('1.0.ica.ICa.nest2', d)
        self.assertIn('1.0.ik1.gK1', d)
        self.assertIn('1.0.ik1.gK1.alpha', d)
        self.assertIn('1.0.ik1.gK1.beta', d)
        self.assertIn('1.0.ik1.IK1', d)
        self.assertIn('1.0.ib.Ib', d)

        # Bound variables
        d = prepare_log(myokit.LOG_BOUND, m, (2, 1))
        self.assertEqual(len(d), 6)
        self.assertIn('0.0.engine.time', d)
        self.assertIn('1.0.engine.time', d)
        self.assertIn('0.0.engine.pace', d)
        self.assertIn('1.0.engine.pace', d)
        self.assertIn('0.0.membrane.i_diff', d)
        self.assertIn('1.0.membrane.i_diff', d)

        # Combinations
        self.assertEqual(
            len(prepare_log(myokit.LOG_NONE + myokit.LOG_BOUND,
                m, (2, 1))), 6)
        self.assertEqual(
            len(prepare_log(myokit.LOG_NONE + myokit.LOG_INTER,
                m, (2, 1))), 64)
        self.assertEqual(
            len(prepare_log(myokit.LOG_BOUND + myokit.LOG_STATE,
                m, (2, 1))), 22)
        self.assertEqual(
            len(prepare_log(myokit.LOG_STATE + myokit.LOG_INTER,
                m, (2, 1))), 80)

        #
        # 1A. Globals and locals
        #
        g = ['engine.time']
        d = prepare_log(myokit.LOG_NONE, m, (2, 1), global_vars=g)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 0)

        # States
        d = prepare_log(myokit.LOG_STATE, m, (2, 1), global_vars=g)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 16)
        self.assertNotIn('engine.time', d)
        d = prepare_log(myokit.LOG_STATE, m, (2, 2), global_vars=g)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 32)
        d = prepare_log(myokit.LOG_STATE, m, (4, 2), global_vars=g)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 64)

        # Internal variables
        d = prepare_log(myokit.LOG_INTER, m, (2, 1), global_vars=g)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 64)

        # Bound variables
        d = prepare_log(myokit.LOG_BOUND, m, (2, 1), global_vars=g)
        self.assertEqual(len(d), 5)
        self.assertIn('engine.time', d)
        self.assertIn('0.0.engine.pace', d)
        self.assertIn('1.0.engine.pace', d)
        self.assertIn('0.0.membrane.i_diff', d)
        self.assertIn('1.0.membrane.i_diff', d)

        # Combinations
        self.assertEqual(
            len(prepare_log(myokit.LOG_NONE + myokit.LOG_BOUND, m, (2, 1),
                global_vars=g)), 5)
        self.assertEqual(
            len(prepare_log(myokit.LOG_NONE + myokit.LOG_INTER, m, (2, 1),
                global_vars=g)), 64)
        self.assertEqual(
            len(prepare_log(myokit.LOG_BOUND + myokit.LOG_STATE, m, (2, 1),
                global_vars=g)), 21)
        self.assertEqual(
            len(prepare_log(myokit.LOG_STATE + myokit.LOG_INTER, m, (2, 1),
                global_vars=g)), 80)

        # Global intermediary variable
        m2 = m.clone()
        x = m2.get('engine').add_variable('x')
        x.set_rhs('5 * pace')
        d = prepare_log(
            myokit.LOG_INTER, m2, (2, 1), global_vars=g + ['engine.x'])
        self.assertIn('engine.x', d)

        #
        # 2. None
        #
        d = prepare_log(None, m, (2, 1), if_empty=myokit.LOG_NONE)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 0)
        d = prepare_log(None, m, (2, 1), if_empty=myokit.LOG_STATE)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 16)

        #
        # 2B. Globals and locals
        #
        d = prepare_log(
            None, m, (2, 1), if_empty=myokit.LOG_NONE, global_vars=g)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 0)
        d = prepare_log(
            None, m, (2, 1), if_empty=myokit.LOG_STATE, global_vars=g)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 16)

        #
        # 3. List of names
        #
        d = prepare_log(('engine.time', 'membrane.V'), m, (2, 1))
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 4)
        d = prepare_log((
            'engine.time',
            'engine.pace',
            'membrane.V',
            'ina.m',
            'ina.h',
            'ina.j',
            'ica.d',
            'ica.f',
            'ik.x',
            'ica.Ca_i',
            'membrane.i_stim',
            'ina.a',
            'ina.m.alpha',
            'ina.m.beta',
            'ina.h.alpha',
            'ina.h.beta',
            'ina.j.alpha',
            'ina.j.beta',
            'ina.INa',
            'ik.xi',
            'ik.x.alpha',
            'ik.x.beta',
            'ik.IK',
            'ikp.IKp',
            'ica.E',
            'ica.d.alpha',
            'ica.d.beta',
            'ica.d.beta.va',
            'ica.d.beta.vb',
            'ica.d.beta.vc',
            'ica.d.beta.vc.vd',
            'ica.f.alpha',
            'ica.f.beta',
            'ica.ICa',
            'ica.ICa.nest1',
            'ica.ICa.nest2',
            'ik1.gK1',
            'ik1.gK1.alpha',
            'ik1.gK1.beta',
            'ik1.IK1',
            'ib.Ib',
        ), m, (2, 1))
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 2 * (31 + 8 + 2))
        self.assertIn('0.0.engine.time', d)
        self.assertIn('0.0.engine.pace', d)
        self.assertIn('0.0.membrane.V', d)
        self.assertIn('0.0.ina.m', d)
        self.assertIn('0.0.ina.h', d)
        self.assertIn('0.0.ina.j', d)
        self.assertIn('0.0.ica.d', d)
        self.assertIn('0.0.ica.f', d)
        self.assertIn('0.0.ik.x', d)
        self.assertIn('0.0.ica.Ca_i', d)
        self.assertIn('0.0.membrane.i_stim', d)
        self.assertIn('0.0.ina.a', d)
        self.assertIn('0.0.ina.m.alpha', d)
        self.assertIn('0.0.ina.m.beta', d)
        self.assertIn('0.0.ina.h.alpha', d)
        self.assertIn('0.0.ina.h.beta', d)
        self.assertIn('0.0.ina.j.alpha', d)
        self.assertIn('0.0.ina.j.beta', d)
        self.assertIn('0.0.ina.INa', d)
        self.assertIn('0.0.ik.xi', d)
        self.assertIn('0.0.ik.x.alpha', d)
        self.assertIn('0.0.ik.x.beta', d)
        self.assertIn('0.0.ik.IK', d)
        self.assertIn('0.0.ikp.IKp', d)
        self.assertIn('0.0.ica.E', d)
        self.assertIn('0.0.ica.d.alpha', d)
        self.assertIn('0.0.ica.d.beta', d)
        self.assertIn('0.0.ica.d.beta.va', d)
        self.assertIn('0.0.ica.d.beta.vb', d)
        self.assertIn('0.0.ica.d.beta.vc', d)
        self.assertIn('0.0.ica.d.beta.vc.vd', d)
        self.assertIn('0.0.ica.f.alpha', d)
        self.assertIn('0.0.ica.f.beta', d)
        self.assertIn('0.0.ica.ICa', d)
        self.assertIn('0.0.ica.ICa.nest1', d)
        self.assertIn('0.0.ica.ICa.nest2', d)
        self.assertIn('0.0.ik1.gK1', d)
        self.assertIn('0.0.ik1.gK1.alpha', d)
        self.assertIn('0.0.ik1.gK1.beta', d)
        self.assertIn('0.0.ik1.IK1', d)
        self.assertIn('0.0.ib.Ib', d)
        self.assertIn('1.0.engine.time', d)
        self.assertIn('1.0.engine.pace', d)
        self.assertIn('1.0.membrane.V', d)
        self.assertIn('1.0.ina.m', d)
        self.assertIn('1.0.ina.h', d)
        self.assertIn('1.0.ina.j', d)
        self.assertIn('1.0.ica.d', d)
        self.assertIn('1.0.ica.f', d)
        self.assertIn('1.0.ik.x', d)
        self.assertIn('1.0.ica.Ca_i', d)
        self.assertIn('1.0.membrane.i_stim', d)
        self.assertIn('1.0.ina.a', d)
        self.assertIn('1.0.ina.m.alpha', d)
        self.assertIn('1.0.ina.m.beta', d)
        self.assertIn('1.0.ina.h.alpha', d)
        self.assertIn('1.0.ina.h.beta', d)
        self.assertIn('1.0.ina.j.alpha', d)
        self.assertIn('1.0.ina.j.beta', d)
        self.assertIn('1.0.ina.INa', d)
        self.assertIn('1.0.ik.xi', d)
        self.assertIn('1.0.ik.x.alpha', d)
        self.assertIn('1.0.ik.x.beta', d)
        self.assertIn('1.0.ik.IK', d)
        self.assertIn('1.0.ikp.IKp', d)
        self.assertIn('1.0.ica.E', d)
        self.assertIn('1.0.ica.d.alpha', d)
        self.assertIn('1.0.ica.d.beta', d)
        self.assertIn('1.0.ica.d.beta.va', d)
        self.assertIn('1.0.ica.d.beta.vb', d)
        self.assertIn('1.0.ica.d.beta.vc', d)
        self.assertIn('1.0.ica.d.beta.vc.vd', d)
        self.assertIn('1.0.ica.f.alpha', d)
        self.assertIn('1.0.ica.f.beta', d)
        self.assertIn('1.0.ica.ICa', d)
        self.assertIn('1.0.ica.ICa.nest1', d)
        self.assertIn('1.0.ica.ICa.nest2', d)
        self.assertIn('1.0.ik1.gK1', d)
        self.assertIn('1.0.ik1.gK1.alpha', d)
        self.assertIn('1.0.ik1.gK1.beta', d)
        self.assertIn('1.0.ik1.IK1', d)
        self.assertIn('1.0.ib.Ib', d)

        #
        # 3B. Globals and local
        #
        d = prepare_log(
            ('engine.time', 'membrane.V'), m, (2, 1), global_vars=g)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 3)
        self.assertIn('engine.time', d)
        self.assertIn('0.0.membrane.V', d)
        self.assertIn('1.0.membrane.V', d)
        d = prepare_log(
            ('engine.time', '0.0.membrane.V'),
            m, (2, 1), global_vars=g)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 2)
        self.assertIn('engine.time', d)
        self.assertIn('0.0.membrane.V', d)
        d = prepare_log(
            ('engine.time', '0.0.membrane.V', '1.0.membrane.V'),
            m, (2, 1), global_vars=g)
        self.assertEqual(len(d), 3)
        self.assertIn('engine.time', d)
        self.assertIn('0.0.membrane.V', d)
        self.assertIn('1.0.membrane.V', d)
        d = prepare_log(
            ('engine.time', '0.0.membrane.V', 'membrane.V'),
            m, (2, 1), global_vars=g)
        self.assertEqual(len(d), 3)
        self.assertIn('engine.time', d)
        self.assertIn('0.0.membrane.V', d)
        self.assertIn('1.0.membrane.V', d)
        d = prepare_log(
            ('engine.time', 'membrane.V', '0.0.membrane.V'),
            m, (2, 1), global_vars=g)
        self.assertEqual(len(d), 3)
        self.assertIn('engine.time', d)
        self.assertIn('0.0.membrane.V', d)
        self.assertIn('1.0.membrane.V', d)

        #
        # 4. Existing log
        #
        # Empty log
        d = prepare_log({}, m)
        self.assertEqual(len(d), 0)
        #
        d = prepare_log({'1.0.engine.time': []}, m, (2, 1))
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 1)
        self.assertIn('1.0.engine.time', d)
        self.assertNotIn('membrane.V', d)
        d = prepare_log({'0.0.engine.time': [1, 2, 3]}, m, (2, 1))
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 1)
        self.assertIn('0.0.engine.time', d)
        self.assertNotIn('0.0.membrane.V', d)
        self.assertNotIn('engine.time', d)
        self.assertEqual(len(d['0.0.engine.time']), 3)
        d = prepare_log(
            {'0.0.engine.time': [1, 2, 3], '1.0.membrane.V': [1, 2, 3]},
            m, (2, 1))
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 2)
        d = prepare_log(d, m, (2, 1))
        self.assertIsInstance(d, myokit.DataLog)
        d = prepare_log(d, m, (2, 1))
        self.assertIsInstance(d, myokit.DataLog)

        #
        # 4B. Existing log, globals and local
        #
        d = prepare_log({'engine.time': [1, 2, 3]}, m, (2, 1), global_vars=g)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 1)
        self.assertIn('engine.time', d)
        self.assertEqual(len(d['engine.time']), 3)
        d = prepare_log(
            {'engine.time': [1, 2, 3], '1.0.membrane.V': [1, 2, 3]},
            m, (2, 1), global_vars=g)
        self.assertIsInstance(d, myokit.DataLog)
        self.assertEqual(len(d), 2)
        d = prepare_log(d, m, (2, 1), global_vars=g)
        self.assertIsInstance(d, myokit.DataLog)
        d = prepare_log(d, m, (2, 1), global_vars=g)
        self.assertIsInstance(d, myokit.DataLog)

        #
        # 5. Errors
        #

        # Unknown global variable
        self.assertRaisesRegex(
            ValueError, 'Unknown variable specified in global', prepare_log,
            myokit.LOG_NONE, m, (2, 1), global_vars=['michael'])

        # State passed as global variable
        self.assertRaisesRegex(
            ValueError, 'State cannot be global', prepare_log, myokit.LOG_NONE,
            m, (2, 1), global_vars=['membrane.V'])

        # Index specified for global variable in log/dict
        self.assertRaisesRegex(
            ValueError, 'index for global', prepare_log,
            {'0.0.engine.time': []}, m, (2, 1), global_vars=['engine.time'])

        # Invalid index for variable in log/dict
        self.assertRaisesRegex(
            ValueError, 'Invalid index', prepare_log, {'3.3.membrane.V': []},
            m, (2, 1))

        # No index for variable that needs it, in log/dict
        self.assertRaisesRegex(
            ValueError, 'non-indexed entry', prepare_log, {'membrane.V': []},
            m, (2, 1))

        # Index specified for global variable in list
        self.assertRaisesRegex(
            ValueError, 'index for global', prepare_log,
            ['0.0.engine.time'], m, (2, 1), global_vars=['engine.time'])

        # Invalid index for variable in list
        self.assertRaisesRegex(
            ValueError, 'Invalid index', prepare_log, ['3.3.membrane.V'], m,
            (2, 1))

    def test_save(self):
        # Test saving in binary format.

        d = myokit.DataLog()
        d['a.b'] = np.arange(0, 100, dtype=np.float32)
        d['c.d'] = np.sqrt(np.arange(0, 100) * 1.2)

        # Test saving with double precision
        with TemporaryDirectory() as td:
            fname = td.path('test.bin')
            d.save(fname)
            e = myokit.DataLog.load(fname)
            self.assertFalse(d is e)
            self.assertEqual(len(d), len(e))
            self.assertIn('a.b', e)
            self.assertIn('c.d', e)
            self.assertNotIn('a.d', e)
            self.assertEqual(len(d['a.b']), len(e['a.b']))
            self.assertEqual(len(d['c.d']), len(e['c.d']))
            self.assertEqual(list(d['a.b']), list(e['a.b']))
            self.assertEqual(list(d['c.d']), list(e['c.d']))
            self.assertTrue(isinstance(d['a.b'][0], np.float32))
            self.assertTrue(isinstance(d['c.d'][0], float))
            self.assertTrue(isinstance(e['a.b'][0], float))
            self.assertTrue(isinstance(e['c.d'][0], float))
            self.assertEqual(e['a.b'].typecode, 'd')
            self.assertEqual(e['c.d'].typecode, 'd')

        # Test saving with single precision
        d = myokit.DataLog()
        d.set_time_key('c.d')
        d['a.b'] = np.arange(0, 100, dtype=np.float32)
        d['c.d'] = np.sqrt(np.arange(0, 100, dtype=np.float32) * 1.2)
        with TemporaryDirectory() as td:
            fname = td.path('test.bin')
            d.save(fname, precision=myokit.SINGLE_PRECISION)
            e = myokit.DataLog.load(fname)
            self.assertFalse(d is e)
            self.assertEqual(len(d), len(e))
            self.assertIn('a.b', e)
            self.assertIn('c.d', e)
            self.assertNotIn('a.d', e)
            self.assertEqual(len(d['a.b']), len(e['a.b']))
            self.assertEqual(len(d['c.d']), len(e['c.d']))
            self.assertEqual(list(d['a.b']), list(e['a.b']))
            self.assertEqual(list(d['c.d']), list(e['c.d']))
            self.assertTrue(isinstance(d['a.b'][0], np.float32))
            self.assertTrue(isinstance(d['c.d'][0], np.float32))
            self.assertTrue(isinstance(e['a.b'][0], float))
            self.assertTrue(isinstance(e['c.d'][0], float))
            self.assertEqual(e['a.b'].typecode, 'f')
            self.assertEqual(e['c.d'].typecode, 'f')
            self.assertEqual(e.time_key(), 'c.d')
            self.assertTrue(np.all(e.time() == d.time()))
            self.assertTrue(np.all(e.time() == d['c.d']))

    def test_load_errors(self):
        # Test if the correct load errors are raised.

        # Missing data file
        path = os.path.join(DIR_IO, 'badlog-1-no-data.zip')
        self.assertRaisesRegex(
            myokit.DataLogReadError, 'log file format', myokit.DataLog.load,
            path)

        # Missing structure file
        path = os.path.join(DIR_IO, 'badlog-2-no-structure.zip')
        self.assertRaisesRegex(
            myokit.DataLogReadError, 'log file format', myokit.DataLog.load,
            path)

        # Not a zip
        path = os.path.join(DIR_IO, 'badlog-3-not-a-zip.zip')
        self.assertRaisesRegex(
            myokit.DataLogReadError, 'bad zip file', myokit.DataLog.load, path)

        # Wrong number of fields
        path = os.path.join(DIR_IO, 'badlog-4-invalid-n-fields.zip')
        self.assertRaisesRegex(
            myokit.DataLogReadError, 'number of fields', myokit.DataLog.load,
            path)

        # Negative data size
        path = os.path.join(DIR_IO, 'badlog-5-invalid-data-size.zip')
        self.assertRaisesRegex(
            myokit.DataLogReadError, 'Invalid data size', myokit.DataLog.load,
            path)

        # Unknown data type
        path = os.path.join(DIR_IO, 'badlog-6-bad-data-type.zip')
        self.assertRaisesRegex(
            myokit.DataLogReadError, 'Invalid data type', myokit.DataLog.load,
            path)

        # Not enough data
        path = os.path.join(DIR_IO, 'badlog-7-not-enough-data.zip')
        self.assertRaisesRegex(
            myokit.DataLogReadError, 'larger data', myokit.DataLog.load, path)

    def test_load_with_progress(self):
        # Test loading with a progress reporter.

        p = TestReporter()
        path = os.path.join(DIR_IO, 'goodlog.zip')
        self.assertFalse(p.entered)
        self.assertFalse(p.exited)
        self.assertFalse(p.updated)
        d = myokit.DataLog.load(path, progress=p)
        self.assertTrue(p.entered)
        self.assertTrue(p.exited)
        self.assertTrue(p.updated)
        self.assertEqual(type(d), myokit.DataLog)

        p = CancellingReporter(1)
        d = myokit.DataLog.load(path, progress=p)
        self.assertIsNone(d)

    def test_save_csv(self):
        # Test saving as csv.

        d = myokit.DataLog()

        # Note: a.b and e.f are both non-decreaing, could be taken for time!
        d['a.b'] = np.arange(0, 100)
        d['c.d'] = np.sqrt(np.arange(0, 100) * 1.2)
        d['e.f'] = np.arange(0, 100) + 1
        with TemporaryDirectory() as td:
            fname = td.path('test.csv')
            d.save_csv(fname)
            e = myokit.DataLog.load_csv(fname)
            self.assertEqual(len(d), len(e))
            self.assertIn('a.b', e)
            self.assertIn('c.d', e)
            self.assertNotIn('a.d', e)
            self.assertEqual(len(d['a.b']), len(e['a.b']))
            self.assertEqual(len(d['c.d']), len(e['c.d']))
            self.assertEqual(list(d['a.b']), list(e['a.b']))
            self.assertEqual(list(d['c.d']), list(e['c.d']))

        # No time key set, will think a.b is time
        self.assertEqual(e.time()[0], d['a.b'][0])

        # Now set time key
        d.set_time_key('a.b')
        with TemporaryDirectory() as td:
            fname = td.path('test.csv')
            d.save_csv(fname)
            e = myokit.DataLog.load_csv(fname)
            self.assertEqual(e.time()[0], d['a.b'][0])

        # Now set time key
        d.set_time_key('e.f')
        with TemporaryDirectory() as td:
            fname = td.path('test.csv')
            d.save_csv(fname)
            e = myokit.DataLog.load_csv(fname)
            self.assertEqual(e.time()[0], d['e.f'][0])

        # Test saving with single precision
        d = myokit.DataLog(time='a.b')
        d['a.b'] = np.arange(0, 100)
        d['c.d'] = np.sqrt(np.arange(0, 100) * 1.2)
        d['e.f'] = np.arange(0, 100) + 1
        d = d.npview()
        with TemporaryDirectory() as td:
            fname = td.path('test.csv')
            d.save_csv(fname, precision=myokit.SINGLE_PRECISION)
            e = myokit.DataLog.load_csv(fname).npview()
            self.assertEqual(len(d), len(e))
            self.assertIn('a.b', e)
            self.assertIn('c.d', e)
            self.assertNotIn('a.d', e)
            self.assertEqual(len(d['a.b']), len(e['a.b']))
            self.assertEqual(len(d['c.d']), len(e['c.d']))
            self.assertTrue(np.all(np.abs(d['a.b'] - e['a.b']) < 1e-6))
            self.assertTrue(np.all(np.abs(d['c.d'] - e['c.d']) < 1e-6))

        # Test saving with python string formats
        d = myokit.DataLog(time='a.b')
        d['a.b'] = np.arange(0, 100)
        d['c.d'] = np.sqrt(np.arange(0, 100) * 1.2)
        d['e.f'] = np.arange(0, 100) + 1
        with TemporaryDirectory() as td:
            fname = td.path('test.csv')
            d.save_csv(fname, precision=None)
            e = myokit.DataLog.load_csv(fname)
            self.assertEqual(len(d), len(e))
            self.assertIn('a.b', e)
            self.assertIn('c.d', e)
            self.assertNotIn('a.d', e)
            self.assertEqual(len(d['a.b']), len(e['a.b']))
            self.assertEqual(len(d['c.d']), len(e['c.d']))
            self.assertTrue(np.all(np.abs(d['a.b'] - e['a.b']) < 1e-6))
            self.assertTrue(np.all(np.abs(d['c.d'] - e['c.d']) < 1e-6))

        # Test invalid precision arguments
        with TemporaryDirectory() as td:
            fname = td.path('test.csv')
            self.assertRaises(ValueError, d.save_csv, fname, 'ernie')

        # Test saving with a fixed order
        d = myokit.DataLog(time='a.b')
        d['a.b'] = np.arange(0, 100)
        d['c.d'] = np.sqrt(np.arange(0, 100) * 1.2)
        d['e.f'] = np.arange(0, 100) + 1
        with TemporaryDirectory() as td:
            fname = td.path('test.csv')
            d.save_csv(fname, order=['a.b', 'c.d', 'e.f'])
            with open(fname, 'r', newline=None) as f:
                header = f.readline()
            self.assertEqual(header, '"a.b","c.d","e.f"\n')

            d.save_csv(fname, order=['e.f', 'a.b', 'c.d'])
            with open(fname, 'r', newline=None) as f:
                header = f.readline()
            self.assertEqual(header, '"e.f","a.b","c.d"\n')

            # Test invalid order
            self.assertRaises(
                ValueError, d.save_csv, fname, order=['e.f', 'a.b', 'c.e'])

        # Test saving with a natural sort order
        d = myokit.DataLog(time='e.t')
        d['e.t'] = np.arange(10)
        d['1.z.a'] = np.arange(10) * 1
        d['0.z.a'] = np.arange(10) * 0
        d['20.z.a'] = np.arange(10) * 20
        d['101.z.a'] = np.arange(10) * 101
        d['100.z.a'] = np.arange(10) * 100
        d['2.z.a'] = np.arange(10) * 2
        d['11.z.a'] = np.arange(10) * 11
        d['14.z.a'] = np.arange(10) * 10
        with TemporaryDirectory() as td:
            fname = td.path('test.csv')
            d.save_csv(fname)
            with open(fname, 'r', newline=None) as f:
                order = [x[1:-1] for x in f.readline().strip().split(',')]
            self.assertEqual(len(order), 9)
            self.assertEqual(order[0], 'e.t')
            self.assertEqual(order[1], '0.z.a')
            self.assertEqual(order[2], '1.z.a')
            self.assertEqual(order[3], '2.z.a')
            self.assertEqual(order[4], '11.z.a')
            self.assertEqual(order[5], '14.z.a')
            self.assertEqual(order[6], '20.z.a')
            self.assertEqual(order[7], '100.z.a')
            self.assertEqual(order[8], '101.z.a')

        # Test saving and loading empty log
        d = myokit.DataLog()
        with TemporaryDirectory() as td:
            fname = td.path('test.csv')
            d.save_csv(fname)
            e = myokit.DataLog.load_csv(fname)
            self.assertEqual(len(e.keys()), 0)

    def test_load_csv_errors(self):
        # Test for errors during csv loading.

        # Test errory file, with comments etc., should work fine!
        path = os.path.join(DIR_IO, 'datalog.csv')
        d = myokit.DataLog.load_csv(path).npview()
        self.assertEqual(set(d.keys()), set(['time', 'v']))
        self.assertTrue(np.all(d['time'] == (1 + np.arange(6))))
        self.assertTrue(np.all(d['v'] == 10 * (1 + np.arange(6))))

        # Empty file
        path = os.path.join(DIR_IO, 'datalog-1-empty.csv')
        d = myokit.DataLog.load_csv(path)
        self.assertEqual(set(d.keys()), set())

        # Test windows line endings
        path = os.path.join(DIR_IO, 'datalog-2-windows.csv')
        d = myokit.DataLog.load_csv(path).npview()
        self.assertEqual(set(d.keys()), set(['time', 'v']))
        self.assertTrue(np.all(d['time'] == (1 + np.arange(6))))
        self.assertTrue(np.all(d['v'] == 10 * (1 + np.arange(6))))

        # Test old mac line endings
        path = os.path.join(DIR_IO, 'datalog-3-old-mac.csv')
        d = myokit.DataLog.load_csv(path).npview()
        self.assertEqual(set(d.keys()), set(['time', 'v']))
        self.assertTrue(np.all(d['time'] == (1 + np.arange(6))))
        self.assertTrue(np.all(d['v'] == 10 * (1 + np.arange(6))))

        # Test empty lines at end
        path = os.path.join(DIR_IO, 'datalog-4-empty-lines.csv')
        d = myokit.DataLog.load_csv(path).npview()
        self.assertEqual(set(d.keys()), set(['time', 'v']))
        self.assertTrue(np.all(d['time'] == (1 + np.arange(6))))
        self.assertTrue(np.all(d['v'] == 10 * (1 + np.arange(6))))

        # Test semicolons at end of each line
        path = os.path.join(DIR_IO, 'datalog-5-semicolons.csv')
        d = myokit.DataLog.load_csv(path).npview()
        self.assertEqual(set(d.keys()), set(['time', 'v']))
        self.assertTrue(np.all(d['time'] == (1 + np.arange(6))))
        self.assertTrue(np.all(d['v'] == 10 * (1 + np.arange(6))))

        # Test unterminated string
        path = os.path.join(DIR_IO, 'datalog-6-open-string.csv')
        self.assertRaisesRegex(
            myokit.DataLogReadError, 'inside quoted', myokit.DataLog.load_csv,
            path)

        # Test empty lines between data
        path = os.path.join(DIR_IO, 'datalog-7-empty-lines-2.csv')
        d = myokit.DataLog.load_csv(path)
        self.assertEqual(set(d.keys()), set(['time', 'v']))
        self.assertTrue(np.all(d['time'] == np.arange(1, 7)))
        self.assertTrue(np.all(d['v'] == 10 * (np.arange(1, 7))))

        # Test unquoted field names
        path = os.path.join(DIR_IO, 'datalog-8-unquoted-header.csv')
        d = myokit.DataLog.load_csv(path).npview()
        self.assertEqual(set(d.keys()), set(['time', 'v']))
        self.assertTrue(np.all(d['time'] == (1 + np.arange(6))))
        self.assertTrue(np.all(d['v'] == 10 * (1 + np.arange(6))))

        # Test double-quoted field names
        path = os.path.join(DIR_IO, 'datalog-9-double-quoted-header.csv')
        d = myokit.DataLog.load_csv(path).npview()
        self.assertEqual(set(d.keys()), set(['time', 'v"quote"']))
        self.assertTrue(np.all(d['time'] == (1 + np.arange(6))))
        self.assertTrue(np.all(d['v"quote"'] == 10 * (1 + np.arange(6))))

        # Test file with just some spaces (one line)
        path = os.path.join(DIR_IO, 'datalog-10-just-spaces.csv')
        d = myokit.DataLog.load_csv(path).npview()
        self.assertEqual(set(d.keys()), set())

        # Test file with just some spaces (one line)
        path = os.path.join(DIR_IO, 'datalog-11-just-a-semicolon.csv')
        d = myokit.DataLog.load_csv(path).npview()
        self.assertEqual(set(d.keys()), set())

        # Test header "abc"x"adc"
        path = os.path.join(DIR_IO, 'datalog-12-bad-header.csv')
        self.assertRaisesRegex(
            myokit.DataLogReadError, 'Expecting double quote',
            myokit.DataLog.load_csv, path)

        # Test empty field "" in header
        path = os.path.join(DIR_IO, 'datalog-13-header-with-empty-1.csv')
        self.assertRaisesRegex(
            myokit.DataLogReadError, 'Empty field', myokit.DataLog.load_csv,
            path)

        # Test empty field "x",,"y" in header
        path = os.path.join(DIR_IO, 'datalog-14-header-with-empty-2.csv')
        self.assertRaisesRegex(
            myokit.DataLogReadError, 'Empty field', myokit.DataLog.load_csv,
            path)

        # Test empty field "time","v", in header
        path = os.path.join(DIR_IO, 'datalog-15-header-with-empty-3.csv')
        self.assertRaisesRegex(
            myokit.DataLogReadError, 'Empty field', myokit.DataLog.load_csv,
            path)

        # Test wrong field count in data
        path = os.path.join(DIR_IO, 'datalog-16-wrong-columns-in-data.csv')
        self.assertRaisesRegex(
            myokit.DataLogReadError, 'Wrong number of columns',
            myokit.DataLog.load_csv, path)

        # Test non-float data
        path = os.path.join(DIR_IO, 'datalog-17-non-float-data.csv')
        self.assertRaisesRegex(
            myokit.DataLogReadError, 'Unable to convert',
            myokit.DataLog.load_csv, path)

    def test_split(self):
        # Test the split function.

        var1 = 'engine.toom'
        var2 = 'membrane.V'
        x = myokit.DataLog({
            var1: [0, 5, 10, 15, 20],
            var2: [1, 2, 3, 4, 5],
        })
        x.set_time_key(var1)

        def split(value):
            s1, s2 = x.split(value)
            return (
                list(s1[var1]),
                list(s2[var1]),
                list(s1[var2]),
                list(s2[var2])
            )

        for i in range(2):
            # First without, then with numpy
            t1, t2, v1, v2 = split(-5)
            self.assertEqual(t1, [])
            self.assertEqual(t2, [0, 5, 10, 15, 20])
            self.assertEqual(v1, [])
            self.assertEqual(v2, [1, 2, 3, 4, 5])
            t1, t2, v1, v2 = split(0)
            self.assertEqual(t1, [])
            self.assertEqual(t2, [0, 5, 10, 15, 20])
            self.assertEqual(v1, [])
            self.assertEqual(v2, [1, 2, 3, 4, 5])
            t1, t2, v1, v2 = split(2)
            self.assertEqual(t1, [0])
            self.assertEqual(t2, [5, 10, 15, 20])
            self.assertEqual(v1, [1])
            self.assertEqual(v2, [2, 3, 4, 5])
            t1, t2, v1, v2 = split(5)
            self.assertEqual(t1, [0])
            self.assertEqual(t2, [5, 10, 15, 20])
            self.assertEqual(v1, [1])
            self.assertEqual(v2, [2, 3, 4, 5])
            t1, t2, v1, v2 = split(7)
            self.assertEqual(t1, [0, 5])
            self.assertEqual(t2, [10, 15, 20])
            self.assertEqual(v1, [1, 2])
            self.assertEqual(v2, [3, 4, 5])
            t1, t2, v1, v2 = split(10)
            self.assertEqual(t1, [0, 5])
            self.assertEqual(t2, [10, 15, 20])
            self.assertEqual(v1, [1, 2])
            self.assertEqual(v2, [3, 4, 5])
            t1, t2, v1, v2 = split(13)
            self.assertEqual(t1, [0, 5, 10])
            self.assertEqual(t2, [15, 20])
            self.assertEqual(v1, [1, 2, 3])
            self.assertEqual(v2, [4, 5])
            t1, t2, v1, v2 = split(15)
            self.assertEqual(t1, [0, 5, 10])
            self.assertEqual(t2, [15, 20])
            self.assertEqual(v1, [1, 2, 3])
            self.assertEqual(v2, [4, 5])
            t1, t2, v1, v2 = split(17)
            self.assertEqual(t1, [0, 5, 10, 15])
            self.assertEqual(t2, [20])
            self.assertEqual(v1, [1, 2, 3, 4])
            self.assertEqual(v2, [5])
            t1, t2, v1, v2 = split(20)
            self.assertEqual(t1, [0, 5, 10, 15])
            self.assertEqual(t2, [20])
            self.assertEqual(v1, [1, 2, 3, 4])
            self.assertEqual(v2, [5])
            t1, t2, v1, v2 = split(22)
            self.assertEqual(t1, [0, 5, 10, 15, 20])
            self.assertEqual(t2, [])
            self.assertEqual(v1, [1, 2, 3, 4, 5])
            self.assertEqual(v2, [])
            x = x.npview()

    def test_split_periodic(self):
        # Test the split_periodic() function.

        tvar = 'engine.toime'
        vvar = 'membrane.V'
        s1 = myokit.DataLog({
            tvar: [0, 500, 999, 1500, 1600, 1700, 2000],
            vvar: [0, 5, 9.99, 15, 16, 17, 20]})
        s1.set_time_key(tvar)
        s2 = myokit.DataLog({
            tvar: [0, 500, 999, 1500, 1600, 1700, 2000, 2001],
            vvar: [0, 5, 9.99, 15, 16, 17, 20, 20.01]})
        s2.set_time_key(tvar)
        n1 = s1.npview()
        n2 = s2.npview()

        # Test default parameters
        #   - endpoint in data
        #   - no split point in data
        logs = s1.split_periodic(1000)
        nlogs = len(logs)
        self.assertEqual(nlogs, 2)
        for var in [tvar, vvar]:
            for log in logs:
                self.assertTrue(var in log)
        self.assertTrue(logs[0][tvar] == [0, 500, 999])
        self.assertTrue(logs[1][tvar] == [1500, 1600, 1700, 2000])

        # Same but now with numpy logs
        logs = n1.split_periodic(1000)
        nlogs = len(logs)
        self.assertEqual(nlogs, 2)
        for var in [tvar, vvar]:
            for log in logs:
                self.assertTrue(var in log)
        self.assertTrue(np.array_equal(logs[0][tvar], np.array([0, 500, 999])))
        self.assertTrue(
            np.array_equal(logs[1][tvar], np.array([1500, 1600, 1700, 2000])))

        # Test default parameters
        #   - no endpoint in data
        #   - split point in data
        logs = s2.split_periodic(1000)
        nlogs = len(logs)
        self.assertEqual(nlogs, 3)
        for var in [tvar, vvar]:
            for log in logs:
                self.assertTrue(var in log)
        self.assertTrue(logs[0][tvar] == [0, 500, 999])
        self.assertTrue(logs[1][tvar] == [1500, 1600, 1700, 2000])
        self.assertTrue(logs[2][tvar] == [2000, 2001])

        # Same but now with numpy logs
        logs = n2.split_periodic(1000)
        nlogs = len(logs)
        self.assertEqual(nlogs, 3)
        for var in [tvar, vvar]:
            for log in logs:
                self.assertTrue(var in log)
        self.assertTrue(np.array_equal(logs[0][tvar], np.array([0, 500, 999])))
        self.assertTrue(np.array_equal(
            logs[1][tvar], np.array([1500, 1600, 1700, 2000])))
        self.assertTrue(np.array_equal(logs[2][tvar], np.array([2000, 2001])))

        # Test with half-closed intervals
        #   - endpoint in data
        #   - no split point in data
        logs = s1.split_periodic(1000, closed_intervals=False)
        nlogs = len(logs)
        self.assertEqual(nlogs, 2)
        for var in [tvar, vvar]:
            for log in logs:
                self.assertTrue(var in log)
        self.assertTrue(logs[0][tvar] == [0, 500, 999])
        self.assertTrue(logs[1][tvar] == [1500, 1600, 1700])

        # Same but now with numpy logs
        logs = n1.split_periodic(1000, closed_intervals=False)
        nlogs = len(logs)
        self.assertEqual(nlogs, 2)
        for var in [tvar, vvar]:
            for log in logs:
                self.assertTrue(var in log)
        self.assertTrue(np.array_equal(logs[0][tvar], np.array([0, 500, 999])))
        self.assertTrue(
            np.array_equal(logs[1][tvar], np.array([1500, 1600, 1700])))

        # Test with half-closed intervals
        #   - no endpoint in data
        #   - split point in data
        logs = s2.split_periodic(1000, closed_intervals=False)
        nlogs = len(logs)
        self.assertEqual(nlogs, 3)
        for var in [tvar, vvar]:
            for log in logs:
                self.assertTrue(var in log)
        self.assertTrue(logs[0][tvar] == [0, 500, 999])
        self.assertTrue(logs[1][tvar] == [1500, 1600, 1700])
        self.assertTrue(logs[2][tvar] == [2000, 2001])

        # Same but now with numpy logs
        logs = n2.split_periodic(1000, closed_intervals=False)
        nlogs = len(logs)
        self.assertEqual(nlogs, 3)
        for var in [tvar, vvar]:
            for log in logs:
                self.assertTrue(var in log)
        self.assertTrue(np.array_equal(logs[0][tvar], np.array([0, 500, 999])))
        self.assertTrue(
            np.array_equal(logs[1][tvar], np.array([1500, 1600, 1700])))
        self.assertTrue(np.array_equal(logs[2][tvar], np.array([2000, 2001])))

        # Test with normalized time
        logs = s2.split_periodic(1000, adjust=True)
        nlogs = len(logs)
        self.assertEqual(nlogs, 3)
        for var in [tvar, vvar]:
            for log in logs:
                self.assertTrue(var in log)
        self.assertTrue(logs[0][tvar] == [0, 500, 999])
        self.assertTrue(logs[1][tvar] == [500, 600, 700, 1000])
        self.assertTrue(logs[2][tvar] == [0, 1])

        # Same but now with numpy logs
        logs = n2.split_periodic(1000, adjust=True)
        nlogs = len(logs)
        self.assertEqual(nlogs, 3)
        for var in [tvar, vvar]:
            for log in logs:
                self.assertTrue(var in log)
        self.assertTrue(np.array_equal(logs[0][tvar], np.array([0, 500, 999])))
        self.assertTrue(
            np.array_equal(logs[1][tvar], np.array([500, 600, 700, 1000])))
        self.assertTrue(np.array_equal(logs[2][tvar], np.array([0, 1])))

        # Test on empty log
        d = myokit.DataLog(time='t')
        self.assertRaises(myokit.InvalidDataLogError, d.split_periodic, 100)
        d['t'] = []
        self.assertRaises(RuntimeError, d.split_periodic, 10)

        # Test negative period
        d['t'] = [1, 2, 3, 4]
        self.assertRaises(ValueError, d.split_periodic, 0)
        self.assertRaises(ValueError, d.split_periodic, -1)

        # Test larger period than log data
        d['x'] = [4, 5, 6, 7]
        e = d.split_periodic(100)
        self.assertEqual(set(d.keys()), set(e.keys()))
        self.assertFalse(d is e)

    def test_trim(self):
        # Test the trim() method.

        d = myokit.DataLog()
        d.set_time_key('t')
        d['t'] = t = [0, 0.1, 0.2, 0.3, 0.4]
        d['a'] = a = [1, 2, 3, 4, 5]
        d['b'] = b = [10, 20, 30, 40, 50]

        # Test function
        def tr(i, j, m, n):
            e = d.trim(i, j)
            self.assertEqual(len(d['t']), 5)
            self.assertEqual(d['t'], t)
            self.assertEqual(d['a'], a)
            self.assertEqual(list(d['b']), list(b))
            self.assertEqual(e['t'], t[m:n])
            self.assertEqual(e['a'], a[m:n])
            self.assertEqual(list(e['b']), list(b[m:n]))

        # Normal operation
        tr(0.1, 0.3, 1, 3)
        tr(-5, 0.3, -5, 3)
        tr(0.2, 30, 2, 30)
        tr(-10, 40, -10, 40)

        # Partial numpy
        d['b'] = b = np.array(b)
        tr(0.1, 0.3, 1, 3)
        tr(-5, 0.3, -5, 3)
        tr(0.2, 30, 2, 30)
        tr(-10, 40, -10, 40)

        # Without adjustment
        d = myokit.DataLog(time='t')
        d['t'] = [1, 2, 3, 4, 5]
        d['v'] = [10, 20, 30, 40, 50]
        e = d.trim(2, 5, adjust=False)
        self.assertEqual(len(e['t']), 3)
        self.assertEqual(len(e['v']), 3)
        self.assertEqual(e['t'], [2, 3, 4])
        self.assertEqual(e['v'], [20, 30, 40])

        # With adjustment
        d = myokit.DataLog(time='t')
        d['t'] = [1, 2, 3, 4, 5]
        d['v'] = [10, 20, 30, 40, 50]
        e = d.trim(2, 5, adjust=True)
        self.assertEqual(len(e['t']), 3)
        self.assertEqual(len(e['v']), 3)
        self.assertEqual(e['t'], [0, 1, 2])
        self.assertEqual(e['v'], [20, 30, 40])

        # Without adjustment, numpy
        d = myokit.DataLog(time='t')
        d['t'] = [1, 2, 3, 4, 5]
        d['v'] = [10, 20, 30, 40, 50]
        d = d.npview()
        e = d.trim(2, 5, adjust=True)
        self.assertEqual(len(e['t']), 3)
        self.assertEqual(len(e['v']), 3)
        self.assertTrue(np.all(e['t'] == [0, 1, 2]))
        self.assertTrue(np.all(e['v'] == [20, 30, 40]))

    def test_trim_left(self):
        # Test the trim_left function.

        var1 = 'engine.toom'
        var2 = 'membrane.V'
        adjust = False
        wnumpy = False

        def trim(value):
            x = myokit.DataLog({
                var1: [0, 5, 10, 15, 20],
                var2: [1, 2, 3, 4, 5],
            })
            x.set_time_key(var1)
            if wnumpy:
                x = x.npview()
            y = x.trim_left(value, adjust=adjust)
            self.assertEqual(len(x[var1]), 5)
            self.assertEqual(len(x[var2]), 5)
            return list(y[var1]), list(y[var2])

        t, v = trim(-5)
        self.assertEqual(t, [0, 5, 10, 15, 20])
        self.assertEqual(v, [1, 2, 3, 4, 5])
        t, v = trim(0)
        self.assertEqual(t, [0, 5, 10, 15, 20])
        self.assertEqual(v, [1, 2, 3, 4, 5])
        t, v = trim(2)
        self.assertEqual(t, [5, 10, 15, 20])
        self.assertEqual(v, [2, 3, 4, 5])
        t, v = trim(5)
        self.assertEqual(t, [5, 10, 15, 20])
        self.assertEqual(v, [2, 3, 4, 5])
        t, v = trim(7)
        self.assertEqual(t, [10, 15, 20])
        self.assertEqual(v, [3, 4, 5])
        t, v = trim(10)
        self.assertEqual(t, [10, 15, 20])
        self.assertEqual(v, [3, 4, 5])
        t, v = trim(13)
        self.assertEqual(t, [15, 20])
        self.assertEqual(v, [4, 5])
        t, v = trim(15)
        self.assertEqual(t, [15, 20])
        self.assertEqual(v, [4, 5])
        t, v = trim(17)
        self.assertEqual(t, [20])
        self.assertEqual(v, [5])
        t, v = trim(20)
        self.assertEqual(t, [20])
        self.assertEqual(v, [5])
        t, v = trim(22)
        self.assertEqual(t, [])
        self.assertEqual(v, [])

        # Now repeat but with adjustments
        adjust = True
        t, v = trim(-5)
        self.assertEqual(t, [5, 10, 15, 20, 25])
        self.assertEqual(v, [1, 2, 3, 4, 5])
        t, v = trim(0)
        self.assertEqual(t, [0, 5, 10, 15, 20])
        self.assertEqual(v, [1, 2, 3, 4, 5])
        t, v = trim(2)
        self.assertEqual(t, [3, 8, 13, 18])
        self.assertEqual(v, [2, 3, 4, 5])
        t, v = trim(5)
        self.assertEqual(t, [0, 5, 10, 15])
        self.assertEqual(v, [2, 3, 4, 5])
        t, v = trim(7)
        self.assertEqual(t, [3, 8, 13])
        self.assertEqual(v, [3, 4, 5])
        t, v = trim(10)
        self.assertEqual(t, [0, 5, 10])
        self.assertEqual(v, [3, 4, 5])
        t, v = trim(13)
        self.assertEqual(t, [2, 7])
        self.assertEqual(v, [4, 5])
        t, v = trim(15)
        self.assertEqual(t, [0, 5])
        self.assertEqual(v, [4, 5])
        t, v = trim(17)
        self.assertEqual(t, [3])
        self.assertEqual(v, [5])
        t, v = trim(20)
        self.assertEqual(t, [0])
        self.assertEqual(v, [5])
        t, v = trim(22)
        self.assertEqual(t, [])
        self.assertEqual(v, [])

        # Now repeat but with numpy and adjustments
        adjust = True
        wnumpy = True
        t, v = trim(-5)
        self.assertEqual(t, [5, 10, 15, 20, 25])
        self.assertEqual(v, [1, 2, 3, 4, 5])
        t, v = trim(0)
        self.assertEqual(t, [0, 5, 10, 15, 20])
        self.assertEqual(v, [1, 2, 3, 4, 5])
        t, v = trim(2)
        self.assertEqual(t, [3, 8, 13, 18])
        self.assertEqual(v, [2, 3, 4, 5])
        t, v = trim(5)
        self.assertEqual(t, [0, 5, 10, 15])
        self.assertEqual(v, [2, 3, 4, 5])
        t, v = trim(7)
        self.assertEqual(t, [3, 8, 13])
        self.assertEqual(v, [3, 4, 5])
        t, v = trim(10)
        self.assertEqual(t, [0, 5, 10])
        self.assertEqual(v, [3, 4, 5])
        t, v = trim(13)
        self.assertEqual(t, [2, 7])
        self.assertEqual(v, [4, 5])
        t, v = trim(15)
        self.assertEqual(t, [0, 5])
        self.assertEqual(v, [4, 5])
        t, v = trim(17)
        self.assertEqual(t, [3])
        self.assertEqual(v, [5])
        t, v = trim(20)
        self.assertEqual(t, [0])
        self.assertEqual(v, [5])
        t, v = trim(22)
        self.assertEqual(t, [])
        self.assertEqual(v, [])

    def test_trim_right(self):
        # Test the trim_right function.

        var1 = 'engine.toom'
        var2 = 'membrane.V'

        def trim(value):
            x = myokit.DataLog({
                var1: [0, 5, 10, 15, 20],
                var2: [1, 2, 3, 4, 5],
            })
            x.set_time_key(var1)
            y = x.trim_right(value)
            self.assertEqual(len(x[var1]), 5)
            self.assertEqual(len(x[var2]), 5)
            return y[var1], y[var2]

        t, v = trim(-5)
        self.assertEqual(t, [])
        self.assertEqual(v, [])
        t, v = trim(0)
        self.assertEqual(t, [])
        self.assertEqual(v, [])
        t, v = trim(2)
        self.assertEqual(t, [0])
        self.assertEqual(v, [1])
        t, v = trim(5)
        self.assertEqual(t, [0])
        self.assertEqual(v, [1])
        t, v = trim(7)
        self.assertEqual(t, [0, 5])
        self.assertEqual(v, [1, 2])
        t, v = trim(10)
        self.assertEqual(t, [0, 5])
        self.assertEqual(v, [1, 2])
        t, v = trim(13)
        self.assertEqual(t, [0, 5, 10])
        self.assertEqual(v, [1, 2, 3])
        t, v = trim(15)
        self.assertEqual(t, [0, 5, 10])
        self.assertEqual(v, [1, 2, 3])
        t, v = trim(17)
        self.assertEqual(t, [0, 5, 10, 15])
        self.assertEqual(v, [1, 2, 3, 4])
        t, v = trim(20)
        self.assertEqual(t, [0, 5, 10, 15])
        self.assertEqual(v, [1, 2, 3, 4])
        t, v = trim(22)
        self.assertEqual(t, [0, 5, 10, 15, 20])
        self.assertEqual(v, [1, 2, 3, 4, 5])

    def test_validate(self):
        # Test the validate() method.

        d = myokit.DataLog()
        d.validate()
        d['time'] = [1, 2, 3]
        d.validate()
        d['time'] = [3, 2, 1]
        d.validate()
        d.set_time_key('time')
        self.assertRaises(myokit.InvalidDataLogError, d.validate)
        d = myokit.DataLog()
        d.set_time_key('time')
        self.assertRaises(myokit.InvalidDataLogError, d.validate)
        d['time'] = [0, 0, 1]
        d.validate()
        d['x'] = [3, 2, 1]
        d.validate()
        self.assertEqual(d.time_key(), 'time')
        self.assertEqual(d.time(), d['time'])
        self.assertEqual(d.time(), [0, 0, 1])
        d['y'] = [4, 5, 6]
        d.validate()
        d['z'] = [7, 8, 9]
        d.validate()
        d['x'].append(1)
        self.assertRaises(myokit.InvalidDataLogError, d.validate)

    def test_apd(self):
        # Test the apd method.

        # Very coarse check
        d = myokit.DataLog(time='time')
        d['time'] = np.linspace(0, 10, 11)
        d['v'] = np.ones(10) * -85
        d['v'][1:3] = 40
        d['v'][6:9] = 40
        apds = d.apd(v='v')
        self.assertEqual(len(apds), 2)
        self.assertTrue(apds['start'][0] > 0)
        self.assertTrue(apds['start'][0] < 1)
        self.assertTrue(apds['duration'][0] > 2.5)
        self.assertTrue(apds['duration'][0] < 3.0)
        self.assertTrue(apds['start'][1] > 5)
        self.assertTrue(apds['start'][1] < 6)
        self.assertTrue(apds['duration'][1] > 3.5)
        self.assertTrue(apds['duration'][1] < 4.0)

        # Check with threshold equal to V
        apds = d.apd(v='v', threshold=-85)
        self.assertEqual(len(apds['start']), 1)
        self.assertEqual(apds['start'][0], 5)
        self.assertEqual(apds['duration'][0], 4)

        # Check against example model
        m, p, x = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        s = myokit.Simulation(m, p)
        s.set_tolerance(1e-8, 1e-8)
        d, apds1 = s.run(
            2000, log=['engine.time', 'membrane.V'],
            log_interval=0.005,
            apd_variable='membrane.V', apd_threshold=-70)
        apds2 = d.apd(threshold=-70)
        self.assertEqual(len(apds1['start']), 2)
        self.assertEqual(len(apds2['start']), 2)
        self.assertAlmostEqual(1, apds1['start'][0] / apds2['start'][0])
        self.assertAlmostEqual(1, apds1['start'][1] / apds2['start'][1])
        self.assertAlmostEqual(1, apds1['duration'][0] / apds2['duration'][0])
        self.assertAlmostEqual(1, apds1['duration'][1] / apds2['duration'][1])

    def test_clone(self):
        # Test data log cloning.

        m, p, x = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        s = myokit.Simulation(m, p)
        d1 = s.run(100, log=myokit.LOG_BOUND + myokit.LOG_STATE).npview()
        d2 = d1.clone()

        # Check keys are the same
        self.assertEqual(d1.keys(), d2.keys())

        # Check the values are the same, but not the same objects
        for k, v in d1.items():
            self.assertTrue(np.all(v == d2[k]))
            self.assertFalse(v is d2[k])
            self.assertTrue(type(d2[k]) == list)

        # Check cloning as numpy arrays
        d2 = d1.clone(numpy=True)
        self.assertEqual(d1.keys(), d2.keys())
        for k, v in d1.items():
            self.assertTrue(np.all(v == d2[k]))
            self.assertFalse(v is d2[k])
            self.assertTrue(type(d2[k]) == np.ndarray)

    def test_fold(self):
        # Test the fold() method.

        d = myokit.DataLog(time='time')
        d['time'] = list(range(100))
        d['x'] = list(np.arange(100) * 3)

        # Test without discarding remainder
        d2 = d.fold(30, discard_remainder=False)
        self.assertEqual(set(d2.keys()), set([
            'time', '0.x', '1.x', '2.x', '3.x']))

        # Test with discarding remainder
        d = d.npview()
        d2 = d.fold(30)
        self.assertEqual(set(d2.keys()), set([
            'time', '0.x', '1.x', '2.x']))
        self.assertEqual(len(d2['time']), 30)
        self.assertEqual(len(d2['0.x']), 30)
        self.assertEqual(len(d2['1.x']), 30)
        self.assertEqual(len(d2['2.x']), 30)
        self.assertTrue(np.all(d2['time'] == d['time'][:30]))
        self.assertTrue(np.all(d2['0.x'] == d['x'][:30]))
        self.assertTrue(np.all(d2['1.x'] == d['x'][30:60]))
        self.assertTrue(np.all(d2['2.x'] == d['x'][60:90]))

    def test_has_nan(self):
        # Test the has_nan() method, which checks if the _final_ value in any
        # field is NaN.

        d = myokit.DataLog(time='time')
        d['time'] = list(range(100))
        d['x'] = list(np.arange(100) * 3)
        d['y'] = list(np.arange(100) * 3)
        d['z'] = list(np.arange(100) * 3)
        self.assertFalse(d.has_nan())
        d['x'][-3] = float('nan')
        self.assertFalse(d.has_nan())
        d['x'][-1] = float('inf')
        self.assertFalse(d.has_nan())
        d['x'][-1] = float('nan')
        self.assertTrue(d.has_nan())

    def test_length(self):
        # Test the length() method, that counts the length of the log's
        # entries.

        d = myokit.DataLog(time='time')
        self.assertEqual(d.length(), 0)
        d['time'] = list(np.arange(100) * 3)
        self.assertEqual(d.length(), 100)
        d['x'] = list(np.arange(100) * 3)
        self.assertEqual(d.length(), 100)

    def test_regularize(self):
        # Test the deprecated regularize() method.

        d = myokit.DataLog(time='time')
        d['time'] = np.log(np.linspace(1, 25, 100))
        d['values'] = np.linspace(1, 25, 100)

        has_scipy = True
        with WarningCollector() as w:
            try:
                e = d.regularize(dt=0.5)
            except ImportError:
                has_scipy = False
        self.assertIn('eprecated', w.text())
        if not has_scipy:
            return

        self.assertEqual(len(e['time']), 7)
        x = np.array([0, 0.5, 1, 1.5, 2, 2.5, 3])
        self.assertTrue(np.all(e['time'] == x))
        for i, y in enumerate(x):
            self.assertTrue(np.abs(np.exp(y) - e['values'][i]) < 0.02)

        # test setting tmin and tmax
        with WarningCollector() as w:
            e = d.regularize(dt=0.5, tmin=0.4, tmax=2.6)
        self.assertEqual(len(e['time']), 5)
        x = np.array([0.4, 0.9, 1.4, 1.9, 2.4])
        self.assertTrue(np.all(e['time'] == x))
        for i, y in enumerate(x):
            self.assertTrue(np.abs(np.exp(y) - e['values'][i]) < 0.02)

    def test_time(self):
        # Test the time() method.

        d = myokit.DataLog(time='t')
        t = [1, 2, 3]
        d['t'] = t
        self.assertIs(d['t'], t)

        # Test non-existing time key
        d = myokit.DataLog(time='t')
        self.assertRaisesRegex(
            myokit.InvalidDataLogError, 'Invalid key', d.time)

        # Test no time key
        d = myokit.DataLog()
        self.assertRaisesRegex(
            myokit.InvalidDataLogError, 'No time', d.time)

    def test_variable_info_errors(self):
        # Test errors raised during variable info checking.

        # Test mismatched dimensions (1d versus 2d)
        d = myokit.DataLog(time='t')
        d['t'] = [1, 2, 3, 4]
        d['0.v'] = [1, 2, 3, 4]
        d['1.1.v'] = [1, 2, 3, 4]
        self.assertRaisesRegex(
            RuntimeError, 'Different dimensions', d.variable_info)
        # Note: Valid log, so not an InvalidDataLogError

        # Test "irregular data": can't be arranged in a rectangular grid
        d = myokit.DataLog(time='t')
        d['t'] = [1, 2, 3, 4]
        d['0.0.v'] = [1, 2, 3, 4]
        d['0.2.v'] = [1, 2, 3, 4]
        d['1.0.v'] = [1, 2, 3, 4]
        d['1.1.v'] = [1, 2, 3, 4]
        self.assertRaisesRegex(RuntimeError, 'Irregular', d.variable_info)
        # Note: Valid log, so not an InvalidDataLogError

    def test_variable_info(self):
        # Test if correct variable info is returned.

        d = myokit.DataLog(time='t')
        # Odd grid
        d['0.0.v'] = [0, 1, 2, 3]
        d['0.2.v'] = [1, 2, 3, 4]
        d['1.0.v'] = [2, 3, 4, 5]
        d['1.2.v'] = [3, 4, 5, 6]
        d['3.0.v'] = [4, 5, 6, 7]
        d['3.2.v'] = [5, 6, 7, 8]
        # Nice grid
        d['0.0.w'] = [0, 1, 2, 3]
        d['0.1.w'] = [1, 2, 3, 4]
        d['1.0.w'] = [2, 3, 4, 5]
        d['1.1.w'] = [3, 4, 5, 6]
        d['2.0.w'] = [4, 5, 6, 7]
        d['2.1.w'] = [5, 6, 7, 8]
        i = d.variable_info()

        # Check found variables
        self.assertEqual(set(i.keys()), set(['v', 'w']))

        # Check if data is in a regular grid
        v = i['v']
        w = i['w']
        self.assertFalse(v.is_regular_grid())
        self.assertTrue(w.is_regular_grid())

        # Check ids
        vids = [(0, 0), (0, 2), (1, 0), (1, 2), (3, 0), (3, 2)]
        self.assertEqual(list(v.ids()), vids)
        wids = [(0, 0), (0, 1), (1, 0), (1, 1), (2, 0), (2, 1)]
        self.assertEqual(list(w.ids()), wids)

        # Check keys
        viter = v.keys()
        for vid in vids:
            self.assertEqual(
                str(vid[0]) + '.' + str(vid[1]) + '.' + 'v', next(viter))
        witer = w.keys()
        for wid in wids:
            self.assertEqual(
                str(wid[0]) + '.' + str(wid[1]) + '.' + 'w', next(witer))

        # Check size
        self.assertEqual(v.size(), (3, 2))
        self.assertEqual(w.size(), (3, 2))

        # Check name
        self.assertEqual(v.name(), 'v')
        self.assertEqual(w.name(), 'w')


if __name__ == '__main__':
    print('Add -v for more debug output')
    import sys
    if '-v' in sys.argv:
        debug = True
    unittest.main()



#
# Provides interaction with the computer algebra system Sympy
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._ereader import SymPyExpressionReader
from ._ewriter import SymPyExpressionWriter


# Importers
# Exporters
# Expression writers
try:
    import sympy
    del sympy
except ImportError:
    _ewriters = {}
else:
    _ewriters = {
        'sympy': SymPyExpressionWriter,
    }


def ewriters():
    """
    Returns a dict of all expression writers available in this module.
    """
    return dict(_ewriters)


# Shared expression reader and writer
_sympyreader_ = None


def _sympyreader(model=None):
    """
    Returns a globally shared SymPyExpressionReader, set to resolve variable
    names using the given model.
    """
    global _sympyreader_
    if _sympyreader_ is None:
        _sympyreader_ = SymPyExpressionReader(model)
    else:
        _sympyreader_.set_model(model)
    return _sympyreader_


_sympywriter_ = None


def _sympywriter():
    """
    Returns a globally shared SymPyExpessionWriter.
    """
    global _sympywriter_
    if _sympywriter_ is None:
        _sympywriter_ = SymPyExpressionWriter()
    return _sympywriter_


# Simple write and read methods
def write(e):
    """
    Converts the given Myokit expression to a SymPy expression.
    """
    w = _sympywriter()
    return w.ex(e)


def read(e, model=None):
    """
    Converts the given Sympy expression to a myokit one. Any variable names
    will be resolved against the given model.
    """
    r = _sympyreader(model)
    return r.ex(e)
#
# Language keywords
#
# None!



#
# Myokit's main module
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
"""
Myokit

This module provides a gateway to the main Myokit components. For example, to
load a model use myokit.load_model('filename.mmt'), create a myokit.Protocol
and then a myokit.Simulation which you can .run() to obtain simulated results.
"""

#__all__ =
#
# __all__ should NOT be provided! Doing so removes all methods below from
# the content imported by "from myokit import *".
#
# Without an explicit __all__, importing * will result in importing all
# functions and classes described below. No submodules of myokit will be
# loaded!
#
# GUI and graphical modules should not be auto-included because they define a
# matplotlib backend to use. If the user requires a different backend, this
# will generate an error.
#

#
# Check python version
#
# Hexversion guide:
#  0x   Hex
#  02   PY_MAJOR_VERSION
#  07   PY_MINOR_VERSION
#  0F   PY_MICRO_VERSION, in hex, so 0F is 15, 10 is 16, etc.
#  F    PY_RELEASE_LEVEL, A for alpha, B for beta, C for candidate, F for final
#  0    PY_RELEASE_SERIAL, increments with every release
#
import sys  # noqa
if sys.hexversion < 0x03000000:  # pragma: no cover
    raise Exception('This version of Myokit does not support Python 2.')
elif sys.hexversion < 0x03070000:  # pragma: no cover
    import warnings  # noqa
    warnings.warn(
        'Myokit is not tested on Python versions before 3.7. Detected'
        f' version {sys.version}.')
    del warnings
del sys


#
# Version information
#
from ._myokit_version import (  # noqa
    __release__,
    __version__,
    __version_tuple__,
)


# Warn about development version
if not __release__:     # pragma: no cover
    import warnings  # noqa
    warnings.warn(f'Using development version of Myokit ({__version__}).')
    del warnings


#
# Licensing
#

# Full license text
LICENSE = """
BSD 3-Clause License

Copyright (c) 2011-2017 Maastricht University. All rights reserved.
Copyright (c) 2017-2020 University of Oxford. All rights reserved.
 (University of Oxford means the Chancellor, Masters and Scholars of the
  University of Oxford, having an administrative office at Wellington Square,
  Oxford OX1 2JD, UK).
Copyright (c) 2020-2023 University of Nottingham. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
""".strip()

# Full license text, html
LICENSE_HTML = """
<h1>Myokit</h1>
<p>
    BSD 3-Clause License
    <br />
    <br />Copyright (c) 2011-2017 Maastricht University. All rights reserved.
    <br />Copyright (c) 2017-2020 University of Oxford. All rights reserved.
    <br />(University of Oxford means the Chancellor, Masters and Scholars of
    the University of Oxford, having an administrative office at Wellington
    Square, Oxford OX1 2JD, UK).
    <br />Copyright (c) 2020-2023 University of Nottingham. All rights
    reserved.</br></p>
<p>
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:
</p>
<ul>
    <li>
        <p>
            Redistributions of source code must retain the above copyright
            notice, this list of conditions and the following disclaimer.
        </p>
    </li>
    <li>
        <p>
            Redistributions in binary form must reproduce the above copyright
            notice, this list of conditions and the following disclaimer in the
            documentation and/or other materials provided with the
            distribution.
        </p>
    </li>
    <li>
        <p>
            Neither the name of the copyright holder nor the names of its
            contributors may be used to endorse or promote products derived
            from this software without specific prior written permission.
        </p>
    </li>
</ul>
<p>
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
    CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
    SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
    INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
    CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
    ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
    POSSIBILITY OF SUCH DAMAGE.
</p>
""".strip()


#
# Paths
#

# Myokit root
import os, inspect  # noqa
try:
    frame = inspect.currentframe()
    DIR_MYOKIT = os.path.abspath(os.path.dirname(inspect.getfile(frame)))
finally:
    # Always manually delete frame
    # https://docs.python.org/2/library/inspect.html#the-interpreter-stack
    del frame

# Binary data files
DIR_DATA = os.path.join(DIR_MYOKIT, '_bin')

# C header files
DIR_CFUNC = os.path.join(DIR_MYOKIT, '_sim')

# Location of myokit user config
DIR_USER = os.path.join(os.path.expanduser('~'), '.config', 'myokit')

# Old user config location: Move if possible
DIR_USER_OLD = os.path.join(os.path.expanduser('~'), '.myokit')

if os.path.exists(DIR_USER_OLD):    # pragma: no cover
    if not os.path.exists(DIR_USER):
        import shutil  # noqa
        shutil.move(DIR_USER_OLD, DIR_USER)
        del shutil

# Ensure the user config directory exists and is writable
if os.path.exists(DIR_USER):    # pragma: no cover
    if not os.path.isdir(DIR_USER):
        raise Exception(
            f'File or link found in place of user directory: {DIR_USER}')
else:                           # pragma: no cover
    os.makedirs(DIR_USER)

# Example mmt file
EXAMPLE = os.path.join(DIR_DATA, 'example.mmt')

# Don't expose standard libraries as part of Myokit
del os, inspect


#
# Debugging modes
#
# Show Generated code, or Write Generated code to file(s):
DEBUG_SG = False
DEBUG_WG = False
# Show compiler output, with lots of warnings enabled:
DEBUG_SC = False
# Show C debug Messages when running compiled code:
DEBUG_SM = False
# Show C Profiling information when running compiled code:
DEBUG_SP = False

#
# Compatibility settings: Some users report problems with output capturing.
#
# Disable capturing (but don't add extra warning flags)
COMPAT_NO_CAPTURE = False
# Disable file-descriptor mode capturing
COMPAT_NO_FD_CAPTURE = False

#
# Data logging flags (bitmasks)
#
LOG_NONE = 0
LOG_STATE = 1
LOG_BOUND = 2
LOG_INTER = 4
LOG_DERIV = 8
LOG_ALL = LOG_STATE + LOG_INTER + LOG_BOUND + LOG_DERIV

#
# Floating point precision
#
SINGLE_PRECISION = 32
DOUBLE_PRECISION = 64

#
# Unit checking modes
#
UNIT_TOLERANT = 1
UNIT_STRICT = 2

#
# Maximum precision float output format strings
#
SFDOUBLE = '{:< 1.17e}'  # Exponent can have 3 digits for very small numbers
SFSINGLE = '{:< 1.9e}'

#
# Date and time formats to use throughout Myokit
#
DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
TIME_FORMAT = '%H:%M:%S'

#
# GUI: Favor PySide or PyQt
#
FORCE_PYQT6 = False
FORCE_PYQT5 = False
FORCE_PYSIDE6 = False
FORCE_PYSIDE2 = False


#
# Library paths and settings
#

# Location of the Sundials (CVODES) shared library objects (.dll or .so)
SUNDIALS_LIB = []

# Location of the Sundials (CVODES) header files (.h)
SUNDIALS_INC = []

# Sundials version number. Defaults to 0.
SUNDIALS_VERSION = 0

# Location of the OpenCL shared library objects (.dll or .so)
OPENCL_LIB = []

# Location of the OpenCL header files (.h)
OPENCL_INC = []


#
# Imports
#

# Exceptions
from ._err import (  # noqa
    CompilationError,
    CyclicalDependencyError,
    DataBlockReadError,
    DataLogReadError,
    DuplicateFunctionArgument,
    DuplicateFunctionName,
    DuplicateName,
    ExportError,
    FindNanError,
    GenerationError,
    IllegalAliasError,
    IllegalReferenceError,
    IllegalReferenceInInitialValueError,
    ImportError,
    IncompatibleModelError,
    IncompatibleUnitError,
    IntegrityError,
    InvalidBindingError,
    InvalidDataLogError,
    InvalidFunction,
    InvalidLabelError,
    InvalidMetaDataNameError,
    InvalidNameError,
    MissingRhsError,
    MissingTimeVariableError,
    MyokitError,
    NumericalError,
    ParseError,
    ProtocolEventError,
    ProtocolParseError,
    SectionNotFoundError,
    SimulationCancelledError,
    SimulationError,
    SimultaneousProtocolEventError,
    UnresolvedReferenceError,
    UnusedVariableError,
    VariableMappingError,
)

# Check if all errors imported
# Dynamically importing them doesn't seem to be possible, and forgetting to
#  import an error creates a hard to debug bug (something needs to go wrong
#  before the interpreter reaches the code raising the error and notices it's
#  not there).
from . import _err  # noqa
import inspect  # noqa
_globals = globals()
ex, name, clas = None, None, None
for ex in inspect.getmembers(_err):
    name, clas = ex
    if type(clas) == type(MyokitError) and issubclass(clas, MyokitError):
        if name not in _globals:    # pragma: no cover
            raise Exception('Failed to import exception: ' + name)
del ex, name, clas, _globals, inspect  # Prevent public visibility
del _err

# Tools
from . import float  # noqa
from . import tools  # noqa

# Model api
from ._model_api import ( # noqa
    check_name,
    Component,
    Equation,
    EquationList,
    Model,
    ModelPart,
    UserFunction,
    Variable,
)

# Expressions
from ._expressions import (  # noqa
    Abs,
    ACos,
    And,
    ASin,
    ATan,
    Ceil,
    Condition,
    Cos,
    Derivative,
    Divide,
    Equal,
    Exp,
    Expression,
    Floor,
    Function,
    If,
    InfixCondition,
    InfixExpression,
    InitialValue,
    Less,
    LessEqual,
    LhsExpression,
    Log,
    Log10,
    Minus,
    More,
    MoreEqual,
    Multiply,
    Name,
    Number,
    Not,
    NotEqual,
    Or,
    PartialDerivative,
    Piecewise,
    Plus,
    Power,
    PrefixCondition,
    PrefixExpression,
    PrefixMinus,
    PrefixPlus,
    Quotient,
    Remainder,
    Sin,
    Sqrt,
    Tan,
)

# Unit and quantity
from ._unit import (  # noqa
    Quantity,
    Unit,
)

# Pacing protocol
from ._protocol import (  # noqa
    PacingSystem,
    Protocol,
    ProtocolEvent,
    TimeSeriesProtocol,
)
from . import pacing  # noqa

# Parser functions
from ._parsing import (  # noqa
    format_parse_error,
    KEYWORDS,
    parse,
    parse_expression_string as parse_expression,
    parse_model,
    parse_protocol,
    parse_state,
    parse_unit_string as parse_unit,
    split,
    strip_expression_units,
)

# Load/save functions
from ._io import (  # noqa
    # Reading, writing
    load,
    load_model,
    load_protocol,
    load_script,
    load_state,
    load_state_bin,
    save,
    save_model,
    save_protocol,
    save_script,
    save_state,
    save_state_bin,
)

# Common units
from . import units  # noqa, also loads all common unit names

# Auxillary functions
from ._aux import (  # noqa
    default_protocol,
    default_script,
    ModelComparison,
    numpy_writer,
    _prepare_bindings,
    python_writer,
    run,
    step,
    version,
    # Deprecated and/or moved to myokit.tools
    Benchmarker,
    date,
    format_float_dict,
    format_path,
    strfloat,
    time,
)

# System information
from ._system import (      # noqa
    system,
)

# Progress reporting
from ._progress import (    # noqa
    ProgressPrinter,
    ProgressReporter,
    Timeout,
)

# Data logging
from ._datalog import (     # noqa
    DataLog,
    _dimco,
    LoggedVariableInfo,
    prepare_log,
    split_key,
)
from ._datablock import (   # noqa
    ColorMap,
    DataBlock1d,
    DataBlock2d,
)

# Simulations
from ._sim import (  # noqa
    CModule,
    CppModule,
    pid_hash,
)
from ._sim.compiler import (  # noqa
    Compiler,
)
from ._sim.sundials import (  # noqa
    Sundials,
)
from ._sim.opencl import (  # noqa
    OpenCL,
    OpenCLDeviceInfo,
    OpenCLInfo,
    OpenCLPlatformInfo,
)
from ._sim.cmodel import CModel             # noqa
from ._sim.cvodessim import Simulation      # noqa
from ._sim.cable import Simulation1d        # noqa
from ._sim.rhs import RhsBenchmarker        # noqa
from ._sim.jacobian import JacobianTracer, JacobianCalculator   # noqa
from ._sim.openclsim import SimulationOpenCL                    # noqa
from ._sim.fiber_tissue import FiberTissueSimulation            # noqa


#
# Globally shared progress reporter
#
_Simulation_progress = None


#
# Load settings
#
from . import _config   # noqa
del _config




#
# Helper function for XML parsing.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#

def split(tag):
    """
    Splits a ``tag`` (as used in elementtree) into a namespace and an element
    name part.
    """
    tag = str(tag)
    if tag[:1] != '{':
        return None, tag
    i = tag.index('}')
    return tag[1:i], tag[1 + i:]




#
# Imports a model from an SBML file.
# Only partial SBML support (Based on SBML level 3 version 2) is provided.
# The file format specifications can be found here
# http://sbml.org/Documents/Specifications.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit.formats


class SBMLImporter(myokit.formats.Importer):
    """
    This :class:`Importer <myokit.formats.Importer>` loads model definitions
    from files in SBML format.

    Not all SBML features are supported, see
    :class:`myokit.formats.sbml.SBMLParser`.
    """
    def __init__(self):
        super().__init__()

    def supports_model(self):
        """Returns ``True``."""
        return True

    def model(self, path):
        """
        Reads the SBML model stored at ``path`` and returns a Myokit model.
        """
        parser = myokit.formats.sbml.SBMLParser()
        model = parser.parse_file(path)
        return model.myokit_model()



#!/usr/bin/env python3
#
# Tests the importers for various formats
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit
import myokit.formats.sympy as mypy


class SymPyReadWriteTest(unittest.TestCase):
    """
    Tests the SymPy ewriter and ereader classes.
    """

    @classmethod
    def setUpClass(cls):
        # Create a model with a variable
        cls._model = myokit.Model()
        component = cls._model.add_component('c')
        bvar = component.add_variable('b')
        bvar.set_rhs(1.23)
        avar = component.add_variable('a')
        avar.set_rhs('b')
        cls._a = myokit.Name(avar)
        cls._b = myokit.Name(bvar)

    def test_reader_writer(self):
        # Test using the proper reader/writer
        try:
            import sympy as sp
        except ImportError:
            print('Sympy not found, skipping test.')
            return

        # Create writer and reader
        w = mypy.SymPyExpressionWriter()
        r = mypy.SymPyExpressionReader(self._model)

        # Name
        a = self._a
        ca = sp.Symbol('c.a')
        self.assertEqual(w.ex(a), ca)
        self.assertEqual(r.ex(ca), a)

        # Number with unit
        b = myokit.Number('12', 'pF')
        cb = sp.Float(12)
        self.assertEqual(w.ex(b), cb)
        # Note: Units are lost in sympy im/ex-port!
        #self.assertEqual(r.ex(cb), b)

        # Number without unit
        b = myokit.Number('12')
        cb = sp.Float(12)
        self.assertEqual(w.ex(b), cb)
        self.assertEqual(r.ex(cb), b)

        # Prefix plus
        x = myokit.PrefixPlus(b)
        self.assertEqual(w.ex(x), cb)
        # Note: Sympy doesn't seem to have a prefix plus
        self.assertEqual(r.ex(cb), b)

        # Prefix minus
        # Note: SymPy treats -x as Mul(NegativeOne, x)
        # But for numbers just returns a number with a negative value
        x = myokit.PrefixMinus(b)
        self.assertEqual(w.ex(x), -cb)
        self.assertEqual(float(r.ex(-cb)), float(x))

        # Plus
        x = myokit.Plus(a, b)
        self.assertEqual(w.ex(x), ca + cb)
        # Note: SymPy likes to re-order the operands...
        self.assertEqual(float(r.ex(ca + cb)), float(x))

        # Minus
        x = myokit.Minus(a, b)
        self.assertEqual(w.ex(x), ca - cb)
        self.assertEqual(float(r.ex(ca - cb)), float(x))

        # Multiply
        x = myokit.Multiply(a, b)
        self.assertEqual(w.ex(x), ca * cb)
        self.assertEqual(float(r.ex(ca * cb)), float(x))

        # Divide
        x = myokit.Divide(a, b)
        self.assertEqual(w.ex(x), ca / cb)
        self.assertEqual(float(r.ex(ca / cb)), float(x))

        # Quotient
        x = myokit.Quotient(a, b)
        self.assertEqual(w.ex(x), ca // cb)
        self.assertEqual(float(r.ex(ca // cb)), float(x))

        # Remainder
        x = myokit.Remainder(a, b)
        self.assertEqual(w.ex(x), ca % cb)
        self.assertEqual(float(r.ex(ca % cb)), float(x))

        # Power
        x = myokit.Power(a, b)
        self.assertEqual(w.ex(x), ca ** cb)
        self.assertEqual(float(r.ex(ca ** cb)), float(x))

        # Sqrt
        x = myokit.Sqrt(a)
        cx = sp.sqrt(ca)
        self.assertEqual(w.ex(x), cx)
        # Note: SymPy converts sqrt to power
        self.assertEqual(float(r.ex(cx)), float(x))

        # Exp
        x = myokit.Exp(a)
        cx = sp.exp(ca)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # Log(a)
        x = myokit.Log(a)
        cx = sp.log(ca)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # Log(a, b)
        x = myokit.Log(a, b)
        cx = sp.log(ca, cb)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(float(r.ex(cx)), float(x))

        # Log10
        x = myokit.Log10(b)
        cx = sp.log(cb, 10)
        self.assertEqual(w.ex(x), cx)
        self.assertAlmostEqual(float(r.ex(cx)), float(x))

        # Sin
        x = myokit.Sin(a)
        cx = sp.sin(ca)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # Cos
        x = myokit.Cos(a)
        cx = sp.cos(ca)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # Tan
        x = myokit.Tan(a)
        cx = sp.tan(ca)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # ASin
        x = myokit.ASin(a)
        cx = sp.asin(ca)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # ACos
        x = myokit.ACos(a)
        cx = sp.acos(ca)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # ATan
        x = myokit.ATan(a)
        cx = sp.atan(ca)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # Floor
        x = myokit.Floor(a)
        cx = sp.floor(ca)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # Ceil
        x = myokit.Ceil(a)
        cx = sp.ceiling(ca)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # Abs
        x = myokit.Abs(a)
        cx = sp.Abs(ca)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # Equal
        x = myokit.Equal(a, b)
        cx = sp.Eq(ca, cb)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # NotEqual
        x = myokit.NotEqual(a, b)
        cx = sp.Ne(ca, cb)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # More
        x = myokit.More(a, b)
        cx = sp.Gt(ca, cb)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # Less
        x = myokit.Less(a, b)
        cx = sp.Lt(ca, cb)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # MoreEqual
        x = myokit.MoreEqual(a, b)
        cx = sp.Ge(ca, cb)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # LessEqual
        x = myokit.LessEqual(a, b)
        cx = sp.Le(ca, cb)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # Not
        x = myokit.Not(a)
        cx = sp.Not(ca)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # And
        cond1 = myokit.More(a, b)
        cond2 = myokit.Less(a, b)
        c1 = sp.Gt(ca, cb)
        c2 = sp.Lt(ca, cb)

        x = myokit.And(cond1, cond2)
        cx = sp.And(c1, c2)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # Or
        x = myokit.Or(cond1, cond2)
        cx = sp.Or(c1, c2)
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # If
        # Note: sympy only does piecewise, not if
        x = myokit.If(cond1, a, b)
        cx = sp.Piecewise((ca, c1), (cb, True))
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x.piecewise())

        # Piecewise
        c = myokit.Number(1)
        cc = sp.Float(1)
        x = myokit.Piecewise(cond1, a, cond2, b, c)
        cx = sp.Piecewise((ca, c1), (cb, c2), (cc, True))
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # Myokit piecewise's (like CellML's) always have a final True
        # condition (i.e. an 'else'). SymPy doesn't require this, so test if
        # we can import this --> It will add an "else 0"
        x = myokit.Piecewise(cond1, a, myokit.Number(0))
        cx = sp.Piecewise((ca, c1))
        self.assertEqual(r.ex(cx), x)

        # SymPy function without Myokit equivalent --> Should raise exception
        cu = sp.principal_branch(cx, cc)
        self.assertRaisesRegex(ValueError, 'Unsupported type', r.ex, cu)

        # Derivative
        m = self._model.clone()
        avar = m.get('c.a')
        r = mypy.SymPyExpressionReader(self._model)
        avar.promote(4)
        x = myokit.Derivative(self._a)
        cx = sp.symbols('dot(c.a)')
        self.assertEqual(w.ex(x), cx)
        self.assertEqual(r.ex(cx), x)

        # Equation
        e = myokit.Equation(a, b)
        ce = sp.Eq(ca, cb)
        self.assertEqual(w.eq(e), ce)
        # There's no backwards equivalent for this!
        # The ereader can handle it, but it becomes and Equals expression.

        # Test sympy division
        del m, avar, x, cx, e, ce
        a = self._model.get('c.a')
        b = self._model.get('c').add_variable('bbb')
        b.set_rhs('1 / a')
        e = b.rhs()
        ce = w.ex(b.rhs())
        e = r.ex(ce)
        self.assertEqual(
            e,
            myokit.Multiply(
                myokit.Number(1),
                myokit.Power(myokit.Name(a), myokit.Number(-1))
            )
        )

        # Test sympy negative numbers
        a = self._model.get('c.a')
        e1 = myokit.PrefixMinus(myokit.Name(a))
        ce = w.ex(e1)
        e2 = r.ex(ce)
        self.assertEqual(e1, e2)

    def test_read_write(self):
        # Test using the read() and write() methods
        try:
            import sympy as sp
        except ImportError:
            print('Sympy not found, skipping test.')
            return

        # Test writing and reading with a model
        a = self._a
        ca = sp.Symbol('c.a')
        self.assertEqual(mypy.write(a), ca)
        self.assertEqual(mypy.read(ca, self._model), a)

        # Test doing it again, with a different model
        m = myokit.Model()
        a = m.add_component('cc').add_variable('aa')
        a = myokit.Name(a)
        ca = sp.Symbol('cc.aa')
        self.assertEqual(mypy.write(a), ca)
        self.assertEqual(mypy.read(ca, m), a)

        # Test reading without a model
        b = myokit.Number('12')
        cb = sp.Float(12)
        self.assertEqual(mypy.write(b), cb)
        self.assertEqual(mypy.read(cb), b)

    def test_access_via_myokit_formats(self):
        # Test access via formats.ewriter.
        try:
            import sympy as sp
        except ImportError:
            print('Sympy not found, skipping test.')
            return
        del sp

        w = myokit.formats.ewriter('sympy')
        self.assertIsInstance(w, mypy.SymPyExpressionWriter)


if __name__ == '__main__':
    unittest.main()
